#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['RemoteNodes', 'RemoteNodesRequest']

from enum import Enum
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from typing import Hashable, Tuple, List, Any, Dict, Callable, Type
from cengal.introspection.inspect import get_exception, entity_owning_module_importable_str, entity_owning_module_info_and_owning_path, entity_properties
from cengal.io.core.memory_management import IOCoreMemoryManagement
from cengal.parallel_execution.asyncio.efficient_streams import StreamManagerIOCoreMemoryManagement, TcpStreamManager, UdpStreamManager, StreamManagerAbstract
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.io.named_connections.named_connections_manager import NamedConnectionsManager
from cengal.code_flow_control.args_manager import number_of_provided_args
from cengal.data_manipulation.serialization import Serializer, Serializers, best_serializer_for_standard_data
from cengal.code_flow_control.args_manager import find_arg_position_and_value, UnknownArgumentError
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from cengal.system import PLATFORM_NAME, PYTHON_VERSION
from importlib import import_module
import sys
import os
import asyncio
import lmdb

from .exceptions import *
from .commands import *
from .class_info import *
from .request_class_info import *
from .remote_node import *
from .serializers import *


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class RemoteNodesRequest(ServiceRequest):
    def start(self, host=None, port=None, tls: bool = True, stream_manager: Optional[StreamManagerAbstract] = None) -> ServiceRequest:
        return self._save(0, host, port)
    def stop(self) -> ServiceRequest:
        return self._save(1)
    def connect(self, host=None, port=None, connection_id_alias=None, tls: bool = True, stream_manager: Optional[StreamManagerAbstract] = None) -> ServiceRequest:
        return self._save(2, host, port, connection_id_alias)
    def disconnect(self, connection_id) -> ServiceRequest:
        return self._save(3, connection_id)


class State(Enum):
    start_initiated = 0
    started = 1
    stop_initiated = 2
    stopped = 3


def is_current_platform(current_platform_name, foreign_platform_name):
    return current_platform_name == foreign_platform_name


class RemoteNodes(DualImmediateProcessingServiceMixin, Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(RemoteNodes, self).__init__(loop)
        self._request_workers = {
            0: self._on_start,
            1: self._on_stop,
            2: self._on_connect,
            2: self._on_disconnect,
        }
        self.platform_name: str = f'{PLATFORM_NAME}-{".".join(PYTHON_VERSION)}'
        self.state: State = State.stopped
        self.servers: Dict[Hashable, RemoteServer] = dict()
        self.clients: Dict[Hashable, RemoteClient] = dict()

        self.serializer__current_platform__custom_types: Serializer = best_serializer_for_standard_data((
            DataFormats.any,
            Tags.current_platform,
            Tags.deep,
            Tags.can_use_custom_types,
            Tags.can_use_bytes,
            Tags.can_use_set,
            Tags.decode_str_as_str,
            Tags.decode_bytes_as_bytes,
            Tags.decode_tuple_as_tuple,
            Tags.decode_list_as_list,
        ), TestDataType.deep_large, 0.1)
        print(self.serializer__current_platform__custom_types.serializer)

        self.serializer__current_platform: Serializer = best_serializer_for_standard_data((
            DataFormats.any,
            Tags.deep,
            Tags.can_use_set,
            Tags.can_use_bytes,
            Tags.decode_str_as_str,
            Tags.decode_bytes_as_bytes,
            Tags.decode_tuple_as_tuple,
            Tags.decode_list_as_list,
        ), TestDataType.deep_large, 0.1)
        print(self.serializer__current_platform.serializer)

        self.serializer__multi_platform: Serializer = best_serializer_for_standard_data((
            DataFormats.any,
            Tags.deep,
            Tags.multi_platform,
            Tags.can_use_bytes,
            Tags.decode_str_as_str,
            Tags.decode_list_as_list,
        ), TestDataType.deep_large, 0.1)
        print(self.serializer__multi_platform.serializer)

        self.serializer__multi_platform_fast: Serializer = best_serializer_for_standard_data((
            DataFormats.any,
            Tags.deep,
            Tags.multi_platform,
            Tags.can_use_bytes,
        ), TestDataType.small, 0.1)
        print(self.serializer__multi_platform_fast.serializer)

        self.serializer__multi_platform__initial_communication: Serializer = best_serializer_for_standard_data((
            DataFormats.json,
            Tags.deep,
            Tags.multi_platform,
        ), TestDataType.small, 0.1)
        print(self.serializer__multi_platform_fast.serializer)
    
    def serialize(self, foreign_platform_name, data) -> Tuple[Serializer, SerializerID, bytes]:
        """_summary_

        Args:
            foreign_platform_name (_type_): _description_
            data (_type_): _description_

        Raises:
            MessageDataCanNotBeSerializedForRequestedNodeError: _description_

        Returns:
            Tuple[Serializer, SerializerID, bytes]: _description_
        """        
        current_platform = is_current_platform(self.platform_name, foreign_platform_name)
        if current_platform:
            serializer_id: SerializerID = SerializerID.current_platform
            serializer: Serializer = self.serializer__current_platform
        else:
            serializer_id = SerializerID.multi_platform
            serializer = self.serializer__multi_platform
        
        try:
            return serializer, serializer_id, serializer.dumps(data)
        except:
            if not current_platform:
                raise MessageDataCanNotBeSerializedForRequestedNodeError
        
        serializer_id = SerializerID.current_platform__custom_types
        serializer = self.serializer__current_platform__custom_types
        return serializer, serializer_id, serializer.dumps(data)
    
    def deserialize(self, serializer_id: int, data: bytes) -> Any:
        """_summary_

        Args:
            serializer_id (int): _description_
            data (bytes): _description_

        Raises:
            UnknownSerializerIDError: _description_

        Returns:
            Any: _description_
        """        
        if SerializerID.multi_platform_fast.value == serializer_id:
            serializer: Serializer = self.serializer__multi_platform_fast
        elif SerializerID.multi_platform.value == serializer_id:
            serializer = self.serializer__multi_platform
        elif SerializerID.current_platform.value == serializer_id:
            serializer = self.serializer__current_platform
        elif SerializerID.current_platform__custom_types.value == serializer_id:
            serializer = self.serializer__current_platform__custom_types
        else:
            raise UnknownSerializerIDError(serializer_id)

        return serializer.loads(data)
    
    def serialize_request(self, server_id: Hashable, args, kwargs) -> bytes:
        """_summary_

        Args:
            server_id (Hashable): _description_
            args (_type_): _description_
            kwargs (_type_): _description_

        Raises:
            MessageCanNotBeEmptyError: _description_
            RuntimeError: _description_
            RuntimeError: _description_

        Returns:
            bytes: _description_
        """        
        result: bytes = None
        remote_server: RemoteServer = self.servers[server_id]
        remotely_registered_service_classes: Dict[Type, LocalClassInfo] = remote_server.remotely_registered_service_classes
        remotely_registered_request_classes: Dict[Type, LocalRequestClassInfo] = remote_server.remotely_registered_request_classes
        try:
            is_raw_request: bool = False
            service_type_param_name: str = 'service_type'
            request_param_name: str  = 'request'
            if 0 == number_of_provided_args(args, kwargs):
                # an error
                raise MessageCanNotBeEmptyError
            elif 2 == number_of_provided_args(args, kwargs):
                service_type: Type[Service] = None
                request_obj: Request = None
                params = {service_type_param_name: 0, request_param_name: 1}
                service_type_param_found, service_type_param_pos, service_type_param_value = find_arg_position_and_value(service_type_param_name, params, args, kwargs)
                if service_type_param_found:
                    if isinstance(service_type_param_value, type) and issubclass(service_type_param_value, Service):
                        service_type = service_type_param_value
                    else:
                        is_raw_request = True
                else:
                    is_raw_request = True
                
                if not is_raw_request:
                    request_param_found, request_param_pos, request_param_value = find_arg_position_and_value(request_param_name, params, args, kwargs)
                    if request_param_found:
                        if isinstance(request_param_value, Request):
                            request_obj = request_param_value
                        else:
                            is_raw_request = True
                    else:
                        is_raw_request = True
                
                if not is_raw_request:
                    if (service_type is None) or (request_obj is None):
                        raise RuntimeError  # executed only if there is some bug in code
                    
                    # check service existance
                    new_service: bool = False
                    if service_type not in remotely_registered_service_classes:
                        new_service = True
                        remote_server.register_service_class(service_type)

                    local_service_class_info: LocalClassInfo = remotely_registered_service_classes[service_type]
                    messages: List[bytes] = list()
                    
                    # register new service
                    if new_service:
                        sys_data: Dict = local_service_class_info()
                        request: Dict = {
                            Fields.command_name.value: Commands.declare_service_class.value,
                            Fields.request_id.value: remote_server.gen_request_id(),
                            Fields.data.value: sys_data,
                        }
                        serialized_request: bytes = self.serializer__multi_platform_fast.dumps(request)
                        messages.append(serialized_request)
                    
                    request_obj_type: Type[Request] = type(request_obj)

                    # check request type existance
                    new_request_type: bool = False
                    if request_obj_type not in remotely_registered_request_classes:
                        new_request_type = True
                        remote_server.register_request_class(request_obj)

                    local_request_class_info: LocalRequestClassInfo = remotely_registered_request_classes[request_obj_type]
                    
                    # register new request type
                    if new_request_type:
                        sys_data: Dict = local_request_class_info()
                        request: Dict = {
                            Fields.command_name.value: Commands.declare_service_request_class.value,
                            Fields.request_id.value: remote_server.gen_request_id(),
                            Fields.data.value: sys_data,
                        }
                        serialized_request: bytes = self.serializer__multi_platform_fast.dumps(request)
                        messages.append(serialized_request)

                    # - actual request to service
                    request_data: Dict = local_request_class_info.request_to_data(request_obj)
                    request_data[CommandDataFieldsServiceRequestWithRequestClass.service_class_id.value] = local_service_class_info.local_id
                    _, serializer_id, request_data = self.serialize(remote_server.foreign_platform_name, request_data)
                    request: Dict = {
                        Fields.command_name.value: Commands.send_service_request_with_request_class.value,
                        Fields.request_id.value: remote_server.gen_request_id(),
                        Fields.is_response_required.value: 1,
                        Fields.data_serializer_id.value: serializer_id.value,
                        Fields.data.value: request_data,
                    }
                    
                    serialized_request: bytes = self.serializer__multi_platform_fast.dumps(request)
                    messages.append(serialized_request)
                    result = b''.join(messages)
            else:
                params = {service_type_param_name: 0}
                found, pos, value = find_arg_position_and_value(service_type_param_name, params, args, kwargs)
                if found:
                    if pos is None:
                        del kwargs[service_type_param_name]
                    else:
                        args = args[1:]
                    
                    if isinstance(value, type) and issubclass(value, Service):
                        service_type: Type[Service] = value

                        # check service existance
                        new_service: bool = False
                        if service_type not in remotely_registered_service_classes:
                            new_service = True
                            remote_server.register_service_class(service_type)

                        local_service_class_info: LocalClassInfo = remotely_registered_service_classes[service_type]
                        
                        messages: List[bytes] = list()
                        
                        # register new service
                        if new_service:
                            sys_data: Dict = local_service_class_info()
                            request: Dict = {
                                Fields.command_name.value: Commands.declare_service_class.value,
                                Fields.request_id.value: remote_server.gen_request_id(),
                                Fields.data.value: sys_data,
                            }
                            serialized_request: bytes = self.serializer__multi_platform_fast.dumps(request)
                            messages.append(serialized_request)

                        # actual request to service
                        request_data: Dict = {
                            CommandDataFieldsServiceRequest.service_class_id.value: local_service_class_info.local_id,
                        }
                        explicit_serializer_required: bool = False
                        if args:
                            explicit_serializer_required = True
                            request_data[CommandDataFieldsServiceRequest.args.value] = args
                        
                        if kwargs:
                            explicit_serializer_required = True
                            request_data[CommandDataFieldsServiceRequest.kwargs.value] = kwargs
                        
                        if explicit_serializer_required:
                            _, serializer_id, request_data = self.serialize(remote_server.foreign_platform_name, request_data)
                        
                        request: Dict = {
                            Fields.command_name.value: Commands.send_service_request.value,
                            Fields.request_id.value: remote_server.gen_request_id(),
                            Fields.is_response_required.value: 1,
                            Fields.data.value: request_data,
                        }
                        if explicit_serializer_required:
                            request[Fields.data_serializer_id.value] = serializer_id.value
                        
                        serialized_request: bytes = self.serializer__multi_platform_fast.dumps(request)
                        messages.append(serialized_request)
                        result = b''.join(messages)
                    else:
                        is_raw_request = True
                else:
                    is_raw_request = True
            
            if is_raw_request:
                # raw request
                request_data: Dict = {
                    CommandDataFieldsRequest.args.value: args,
                    CommandDataFieldsRequest.kwargs.value: kwargs,
                }
                _, serializer_id, request_data = self.serialize(remote_server.foreign_platform_name, request_data)
                
                request: Dict = {
                    Fields.command_name.value: Commands.send_request.value,
                    Fields.request_id.value: remote_server.gen_request_id(),
                    Fields.is_response_required.value: 1,
                    Fields.data_serializer_id.value: serializer_id.value,
                    Fields.data.value: request_data,
                }
                
                result = self.serializer__multi_platform_fast.dumps(request)
        except UnknownArgumentError:
            raise RuntimeError  # executed only if there is some bug in code
        
        return result

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'instances num': len(self.instances),
        }

    def single_task_registration_or_immediate_processing_single(
            self, *args, **kwargs
    ) -> Tuple[bool, Optional[CoroID], Any]:
        number_of_args: int = number_of_provided_args(args, kwargs)
        if 1 == number_of_args:
            return self._on_get(*args, **kwargs)
        elif 2 == number_of_args:
            return self._on_set(*args, **kwargs)
        else:
            return True, None, RuntimeError(f'Wrong number of parameters: {number_of_args}')

    def full_processing_iteration(self):
        results_bak = self.results
        self.results = type(results_bak)()
        for waiter_coro_id, instance in results_bak.items():
            self.register_response(waiter_coro_id, instance)

        self.make_dead()

    def in_work(self) -> bool:
        return self.thrifty_in_work(bool(self.results))
    
    def _on_start(self, host=None, port=None):
        if State.stopped != self.state:
            return True, None, RuntimeError('Already started or not yet stopped.')
        else:
            ...
        
        return True, None, None


RemoteNodesRequest.default_service_type = RemoteNodes



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['Sleep']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.time_management.timer import Timer
from functools import partial
from typing import Optional, Tuple, Union


class Sleep(TypedService[float]):
    def __init__(self, loop: CoroSchedulerType):
        super(Sleep, self).__init__(loop)
        self.timer = Timer()
        self.pending_tasks_number = 0
        self.pending_foreground_tasks_number = 0

    def single_task_registration_or_immediate_processing(
            self, delay: float) -> Tuple[bool, None, None]:
        def timer_handler_func(coro_id: CoroID, foreground: bool, start_time: float):
            current_time = perf_counter()
            if start_time > current_time:
                start_time = current_time
            real_delay = current_time - start_time
            self.task_triggered(foreground)
            self.register_response(coro_id, real_delay, None)

        foreground: bool = not self.current_caller_coro_info.coro.is_background_coro
        handler = partial(timer_handler_func, self.current_caller_coro_info.coro_id, foreground, perf_counter())
        self.task_added(foreground)
        self.timer.register(handler, delay)
        self.make_live()
        return False, None, None

    def full_processing_iteration(self):
        self.timer()
        if 0 == self.pending_tasks_number:
            self.make_dead()

    def task_added(self, foreground: bool):
        self.pending_tasks_number += 1
        self.pending_foreground_tasks_number += 1 if foreground else 0

    def task_triggered(self, foreground: bool):
        self.pending_tasks_number -= 1
        self.pending_foreground_tasks_number -= 1 if foreground else 0

    def in_work(self) -> bool:
        return self.thrifty_in_work(self.pending_tasks_number != 0)
    
    def in_forground_work(self) -> bool:
        return self.pending_foreground_tasks_number
    
    def time_left_before_next_event(self) -> Tuple[bool, Optional[Union[int, float]]]:
        return self.pending_tasks_number != 0, self.timer.nearest_event()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['find_free_port', 'simple_port_search', 'asimple_port_search', 'find_free_ports_range', 'simple_ports_range_search', 'asimple_ports_range_search']


import asyncio
from cengal.io.used_ports import used_ports, PortsIterator, Protocol, PortStatus, UsedPorts
from cengal.math.numbers import RationalNumber
from typing import Optional, Union, Tuple, Set
from gc import collect


def client_connected_cb(*args, **kwargs):
    pass


async def find_free_port(host, ports_iterartor: PortsIterator, timeout: RationalNumber = 0) -> Optional[int]:
    """_summary_
    Example:

    from cengal.io.serve_free_ports import find_free_port
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports
    import asyncio

    async def main():
        ports_rages_iterator: PortsIterator = used_ports().range(Protocol.tcp, slice(18000, 18100), {PortStatus.na, PortStatus.no})
        port: int = await find_free_port('0.0.0.0', ports_rages_iterator, 5.0)
        if port is not None:
            print(port)  # Output: 18000

    Args:
        host (_type_): _description_
        ports_iterartor (PortsIterator): _description_
        timeout (RationalNumber, optional): _description_. Defaults to 0.

    Returns:
        Optional[int]: _description_
    """    
    for ports_range in ports_iterartor:
        for port in range(ports_range.start, ports_range.stop):
            try:
                if timeout:
                    server = await asyncio.wait_for(asyncio.start_server(client_connected_cb, host, port), timeout)
                else:
                    server = await asyncio.start_server(client_connected_cb, host, port)
                
                try:
                    server.close()
                except Exception:
                    pass
                finally:
                    del server
                    collect()

                return port
            except TimeoutError as err:
                pass
            except (ConnectionRefusedError, ConnectionError, OSError) as err:
                pass


def simple_port_search(host: str, port_or_range: Union[int, slice, Tuple[int, int]], protocol: Protocol = Protocol.tcp, statuses: Optional[Union[PortStatus, Set[PortStatus]]] = None, timeout: RationalNumber = 0, used_ports_instance: Optional[UsedPorts] = None) -> Optional[int]:
    """_summary_

    Example:

    from cengal.io.serve_free_ports import simple_port_search
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports

    def main():
        port: int = simple_port_search('0.0.0.0', slice(18000, 18100), Protocol.tcp, {PortStatus.na, PortStatus.no}, 5.0)
        if port is not None:
            print(port)  # Output: 18000

    Args:
        host (str): _description_
        port_or_range (Union[int, slice, Tuple[int, int]]): _description_
        protocol (Protocol, optional): _description_. Defaults to Protocol.tcp.
        statuses (Optional[Union[PortStatus, Set[PortStatus]]], optional): _description_. Defaults to None.
        desired_number_of_ports (int, optional): _description_. Defaults to 1.
        timeout (int, optional): _description_. Defaults to 0.
        used_ports_instance (Optional[UsedPorts], optional): .

    Returns:
        _type_: _description_
    """    
    if statuses is None: statuses = {PortStatus.na, PortStatus.no}
    if used_ports_instance is None: used_ports_instance = used_ports()
    return asyncio.run(find_free_port(host, used_ports_instance.port(protocol, port_or_range, statuses), timeout))


async def asimple_port_search(host: str, port_or_range: Union[int, slice, Tuple[int, int]], protocol: Protocol = Protocol.tcp, statuses: Optional[Union[PortStatus, Set[PortStatus]]] = None, timeout: RationalNumber = 0, used_ports_instance: Optional[UsedPorts] = None) -> Optional[int]:
    """_summary_

    Example:

    from cengal.io.serve_free_ports import asimple_port_search
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports
    import asyncio

    async def main():
        port: int = await asimple_port_search('0.0.0.0', slice(18000, 18100), Protocol.tcp, {PortStatus.na, PortStatus.no}, 5.0)
        if port is not None:
            print(port)  # Output: 18000

    Args:
        host (str): _description_
        port_or_range (Union[int, slice, Tuple[int, int]]): _description_
        protocol (Protocol, optional): _description_. Defaults to Protocol.tcp.
        statuses (Optional[Union[PortStatus, Set[PortStatus]]], optional): _description_. Defaults to None.
        desired_number_of_ports (int, optional): _description_. Defaults to 1.
        timeout (int, optional): _description_. Defaults to 0.
        used_ports_instance (Optional[UsedPorts], optional): .

    Returns:
        _type_: _description_
    """    
    if statuses is None: statuses = {PortStatus.na, PortStatus.no}
    if used_ports_instance is None: used_ports_instance = used_ports()
    return await find_free_port(host, used_ports_instance.port(protocol, port_or_range, statuses), timeout)


async def find_free_ports_range(host, ports_iterartor: PortsIterator, timeout: RationalNumber = 0) -> Optional[slice]:
    """
    Example:

    from cengal.io.serve_free_ports import find_free_ports_range
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports
    import asyncio

    async def main():
        ports_rages_iterator: PortsIterator = used_ports().range(Protocol.tcp, slice(18000, 18100), {PortStatus.na, PortStatus.no}, 3)
        ports_range = await find_free_ports_range('0.0.0.0', ports_rages_iterator, 5.0)
        if ports_range is not None:
            first_port = ports_range.start  # Output: 18000
            last_port = ports_range.stop - 1  # Output: 18002
            first_port, last_port = unify_ports(ports_range)  # Output: (18000, 18002)
            for port in range(ports_range.start, ports_range.stop):
                print(port)
            
            print(list(range(ports_range.start, ports_range.stop)))  # Output: [18000, 18001, 18002]
            if isinstance(purify_ports(ports_range), slice):
                # type is `slice`
                print('Ports range')
            else:
                # type is `int`
                print('Single port')

    Args:
        host (_type_): _description_
        ports_iterartor (PortsIterator): _description_
        timeout (int, optional): _description_. Defaults to 0.

    Returns:
        Optional[slice]: _description_
    """
    for ports_range in ports_iterartor:
        try:
            for port in range(ports_range.start, ports_range.stop):
                if timeout:
                    server = await asyncio.wait_for(asyncio.start_server(client_connected_cb, host, port), timeout)
                else:
                    server = await asyncio.start_server(client_connected_cb, host, port)
                
                try:
                    server.close()
                except Exception:
                    pass
                finally:
                    del server
                    collect()

            return ports_range
        except TimeoutError as err:
            pass
        except (ConnectionRefusedError, ConnectionError, OSError) as err:
            pass


def simple_ports_range_search(host: str, port_or_range: Union[int, slice, Tuple[int, int]], protocol: Protocol = Protocol.tcp, statuses: Optional[Union[PortStatus, Set[PortStatus]]] = None, desired_number_of_ports: int = 1, timeout: RationalNumber = 0, used_ports_instance: Optional[UsedPorts] = None) -> Optional[slice]:
    """_summary_

    Example:

    from cengal.io.serve_free_ports import simple_ports_range_search
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports

    def main():
        ports_range = simple_ports_range_search('0.0.0.0', slice(18000, 18100), Protocol.tcp, {PortStatus.na, PortStatus.no}, 3, 5.0)
        if ports_range is not None:
            first_port = ports_range.start  # Output: 18000
            last_port = ports_range.stop - 1  # Output: 18002
            first_port, last_port = unify_ports(ports_range)  # Output: (18000, 18002)
            for port in range(ports_range.start, ports_range.stop):
                print(port)
            
            print(list(range(ports_range.start, ports_range.stop)))  # Output: [18000, 18001, 18002]
            if isinstance(purify_ports(ports_range), slice):
                # type is `slice`
                print('Ports range')
            else:
                # type is `int`
                print('Single port')

    Args:
        host (str): _description_
        port_or_range (Union[int, slice, Tuple[int, int]]): _description_
        protocol (Protocol, optional): _description_. Defaults to Protocol.tcp.
        statuses (Optional[Union[PortStatus, Set[PortStatus]]], optional): _description_. Defaults to None.
        desired_number_of_ports (int, optional): _description_. Defaults to 1.
        timeout (int, optional): _description_. Defaults to 0.
        used_ports_instance (Optional[UsedPorts], optional): .

    Returns:
        _type_: _description_
    """    
    if statuses is None: statuses = {PortStatus.na, PortStatus.no}
    if used_ports_instance is None: used_ports_instance = used_ports()
    return asyncio.run(find_free_ports_range(host, used_ports_instance.range(protocol, port_or_range, statuses, desired_number_of_ports), timeout))


async def asimple_ports_range_search(host: str, port_or_range: Union[int, slice, Tuple[int, int]], protocol: Protocol = Protocol.tcp, statuses: Optional[Union[PortStatus, Set[PortStatus]]] = None, desired_number_of_ports: int = 1, timeout: RationalNumber = 0, used_ports_instance: Optional[UsedPorts] = None) -> Optional[slice]:
    """_summary_

    Example:

    from cengal.io.serve_free_ports import asimple_ports_range_search
    from cengal.io.used_ports import PortsIterator, used_ports, Protocol, PortStatus, unify_ports, purify_ports
    import asyncio

    async def main():
        ports_range = await asimple_ports_range_search('0.0.0.0', slice(18000, 18100), Protocol.tcp, {PortStatus.na, PortStatus.no}, 3, 5.0)
        if ports_range is not None:
            first_port = ports_range.start  # Output: 18000
            last_port = ports_range.stop - 1  # Output: 18002
            first_port, last_port = unify_ports(ports_range)  # Output: (18000, 18002)
            for port in range(ports_range.start, ports_range.stop):
                print(port)
            
            print(list(range(ports_range.start, ports_range.stop)))  # Output: [18000, 18001, 18002]
            if isinstance(purify_ports(ports_range), slice):
                # type is `slice`
                print('Ports range')
            else:
                # type is `int`
                print('Single port')

    Args:
        host (str): _description_
        port_or_range (Union[int, slice, Tuple[int, int]]): _description_
        protocol (Protocol, optional): _description_. Defaults to Protocol.tcp.
        statuses (Optional[Union[PortStatus, Set[PortStatus]]], optional): _description_. Defaults to None.
        desired_number_of_ports (int, optional): _description_. Defaults to 1.
        timeout (int, optional): _description_. Defaults to 0.
        used_ports_instance (Optional[UsedPorts], optional): .

    Returns:
        _type_: _description_
    """    
    if statuses is None: statuses = {PortStatus.na, PortStatus.no}
    if used_ports_instance is None: used_ports_instance = used_ports()
    return await find_free_ports_range(host, used_ports_instance.range(protocol, port_or_range, statuses, desired_number_of_ports), timeout)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from collections import deque
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence

"""
Динамический кусочный массив данных. Заполняется отдельными массивами байт (например принятых по TCP во время
ассинхронного режима работы сокета), но функционирует так, как буд-то это цельный непрерывный массив.
Дает значительный выигрыш в скорости по сравнению с обновлением встроенного контейнера list() за счет гораздо
меньшего количества аллокаций и копирований памяти.
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


if cython.compiled:
    # print('"{}" - compiled.'.format(__name__))
    pass
else:
    print('"{}" - interpreted script.'.format(__name__))


cdef class DynamicListOfPieces:
    cdef object _init_param__joiner
    cdef long long _init_param__on_hold_limit

    cdef object _joiner

    cdef object _data
    cdef long long _data_length

    cdef long long _offset_limit
    cdef long long _useful_size
    cdef long long _real_size

    def __init__(self, object joiner=None, long long on_hold_limit=0):
        self._init_param__joiner = joiner
        self._init_param__on_hold_limit = on_hold_limit

        self._joiner = b'' if joiner is None else joiner

        self._data = list()
        self._data_length = 0

        self._offset_limit = on_hold_limit or 1000
        self._useful_size = 0
        self._real_size = 0

    def add_piece_of_data(self, object piece_of_data):
        self._data.append(piece_of_data)
        self._data_length += len(piece_of_data)
        self._useful_size += 1
        self._real_size += 1

    def size(self):
        return self._data_length

    def get_data(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long result_data_length = 0
        cdef long long pieces_qnt = self._real_size - self._useful_size
        cdef long long delta_range = pieces_qnt

        cdef object result = None

        cdef object piece
        cdef long long piece_length = 0

        cdef object result_list = None

        cdef object splittable_piece = None
        cdef long long needed = 0
        cdef object last_current_data_part = None
        cdef object first_undecoded_data_part = None

        if size < len(self._data[pieces_qnt]):
            result = self._data[pieces_qnt][:size]
            self._data_length -= size
            self._data[pieces_qnt] = self._data[pieces_qnt][size:]
            return result
        elif size == len(self._data[pieces_qnt]):
            result = self._data[pieces_qnt]
            self._data_length -= size
            pieces_qnt += 1
            self._useful_size -= 1
            if pieces_qnt >= self._offset_limit:
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]
            return result
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                # if (len(splittable_piece) > 4096 * 1024**2) and (type(splittable_piece) is not memoryview):
                #     splittable_piece = memoryview(splittable_piece)
                needed = size - result_data_length

                last_current_data_part = splittable_piece[:needed]
                first_undecoded_data_part = splittable_piece[needed:]

                # result_list.append(last_current_data_part)
                self._data[pieces_qnt] = first_undecoded_data_part

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(last_current_data_part)

            else:
                result_list = self._data[delta_range:pieces_qnt]

            self._useful_size = self._real_size - pieces_qnt
            if pieces_qnt >= self._offset_limit:
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]

            self._data_length -= size

            result = self._joiner.join(result_list)

            if len(result) > 4096:
                result = memoryview(result)

            return result

    def read_data(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long result_data_length = 0
        cdef long long pieces_qnt = self._real_size - self._useful_size
        cdef long long delta_range = pieces_qnt

        cdef object piece
        cdef long long piece_length = 0

        cdef object result_list = None

        cdef object splittable_piece = None
        cdef long long needed = 0

        cdef object result

        if size < len(self._data[pieces_qnt]):
            return self._data[pieces_qnt][:size]
        elif size == len(self._data[pieces_qnt]):
            return self._data[pieces_qnt]
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                needed = size - result_data_length

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(splittable_piece[:needed])

            else:
                result_list = self._data[delta_range:pieces_qnt]

            result = self._joiner.join(result_list)

            if len(result) > 4096:
                result = memoryview(result)

            return result

    def remove(self):
        pass

    def __copy__(self):
        return DynamicListOfPieces(self._init_param__joiner,
                                   self._init_param__on_hold_limit)


cdef class DynamicListOfPiecesWithLengthControl(DynamicListOfPieces):
    cdef bint _removed

    cdef long long _init_param__on_hold_data_size_limit
    cdef object _init_param__external_data_length
    cdef object _init_param__external_deletable_data_full_size

    cdef long long _on_hold_data_size_limit
    cdef long long _deletable_data_full_size

    cdef object _external_data_length
    cdef object _external_deletable_data_full_size

    def __init__(self, object joiner=None, long long on_hold_limit=0, long long on_hold_data_size_limit=0,
                 object external_data_length=None,
                 object external_deletable_data_full_size=None):
        super(DynamicListOfPiecesWithLengthControl, self).__init__(joiner, on_hold_limit)

        self._removed = True

        self._init_param__on_hold_data_size_limit = on_hold_data_size_limit
        self._init_param__external_data_length = external_data_length
        self._init_param__external_deletable_data_full_size = external_deletable_data_full_size

        self._on_hold_data_size_limit = on_hold_data_size_limit or 1024**2
        self._deletable_data_full_size = 0

        self._external_data_length = external_data_length or ResultExistence(False, 0)
        self._external_deletable_data_full_size = external_deletable_data_full_size or ResultExistence(False, 0)

    def add_piece_of_data(self, object piece_of_data):
        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._external_data_length.result += piece_of_data_len
        self._useful_size += 1
        self._real_size += 1

    # @profile
    def get_data(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long result_data_length = 0
        cdef long long pieces_qnt = self._real_size - self._useful_size
        cdef long long delta_range = pieces_qnt

        cdef object result = None

        cdef object piece
        cdef long long piece_length = 0

        cdef object result_list = None

        cdef object splittable_piece = None
        cdef long long needed = 0
        cdef object last_current_data_part = None
        cdef object first_undecoded_data_part = None

        first_piece_len = len(self._data[pieces_qnt])
        if size < first_piece_len:
            result = self._data[pieces_qnt][:size]
            self._data_length -= size
            self._external_data_length.result -= size
            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            self._data[pieces_qnt] = self._data[pieces_qnt][size:]
            return result
        elif size == first_piece_len:
            result = self._data[pieces_qnt]
            self._data_length -= size
            self._external_data_length.result -= size
            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            pieces_qnt += 1
            self._useful_size -= 1
            if (pieces_qnt >= self._offset_limit) or (self._deletable_data_full_size > self._on_hold_data_size_limit):
                self._external_deletable_data_full_size.result -= self._deletable_data_full_size
                self._deletable_data_full_size = 0
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]
            return result
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                # if (len(splittable_piece) > 4096 * 1024**2) and (type(splittable_piece) is not memoryview):
                #     splittable_piece = memoryview(splittable_piece)
                needed = size - result_data_length

                last_current_data_part = splittable_piece[:needed]
                first_undecoded_data_part = splittable_piece[needed:]

                # result_list.append(last_current_data_part)
                self._data[pieces_qnt] = first_undecoded_data_part

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(last_current_data_part)

            else:
                result_list = self._data[delta_range:pieces_qnt]

            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            self._useful_size = self._real_size - pieces_qnt
            if pieces_qnt >= self._offset_limit:
                self._external_deletable_data_full_size.result -= self._deletable_data_full_size
                self._deletable_data_full_size = 0
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]

            self._data_length -= size
            self._external_data_length.result -= size

            # result = self._joiner.join(result_list)
            #
            # if len(result) > 4096:
            #     result = memoryview(result)
            #
            # return result

            return memoryview(self._joiner.join(result_list))

    def remove(self):
        if not self._removed:
            self._external_data_length.result -= self._data_length
            self._external_deletable_data_full_size.result -= self._deletable_data_full_size
            self._removed = True

    def __copy__(self):
        return DynamicListOfPiecesWithLengthControl(self._init_param__joiner,
                                                    self._init_param__on_hold_limit,
                                                    self._init_param__on_hold_data_size_limit,
                                                    self._init_param__external_data_length,
                                                    self._init_param__external_deletable_data_full_size)

    def __del__(self):
        self.remove()


cdef class DynamicListOfPiecesDequeWithLengthControl:
    cdef bint _removed

    cdef object _init_param__joiner
    cdef object _init_param__external_data_length

    cdef object _joiner

    cdef object _data
    cdef long long _data_length

    cdef long long _real_size

    cdef object _external_data_length

    def __init__(self, object joiner=None, object external_data_length=None):
        self._removed = True

        self._init_param__joiner = joiner
        self._init_param__external_data_length = external_data_length

        self._joiner = b'' if joiner is None else joiner
        self._data = deque()
        self._data_length = 0

        self._real_size = 0

        self._external_data_length = external_data_length or ResultExistence(False, 0)

    def size(self):
        return self._data_length

    def qnt(self):
        return self._real_size

    def add_piece_of_data(self, object piece_of_data):
        cdef long long piece_of_data_len

        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._external_data_length.result += piece_of_data_len
        self._real_size += 1

    # @profile
    def get_data(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long first_piece_len
        cdef object current_piece
        cdef object result_data
        cdef long long result_size
        cdef object another_piece
        cdef long long another_piece_len
        cdef long long delta_size

        first_piece_len = len(self._data[0])
        if size < first_piece_len:
            current_piece = memoryview(self._data[0])
            self._data[0] = current_piece[size:]
            self._data_length -= size
            self._external_data_length.result -= size
            return current_piece[:size]
        elif size == first_piece_len:
            self._data_length -= size
            self._external_data_length.result -= size
            self._real_size -= 1
            # return memoryview(self._data.popleft())
            return self._data.popleft()
        else:
            result_data = deque()
            result_size = 0

            while self._real_size and (result_size < size):
                another_piece = self._data.popleft()
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                self._data_length -= another_piece_len
                self._external_data_length.result -= another_piece_len
                self._real_size -= 1
                result_size += another_piece_len
                if (result_size < size) and ((len(self._data[0]) + result_size) > size):
                    break

            if result_size < size:
                delta_size = size - result_size
                current_piece = memoryview(self._data[0])
                result_data.append(current_piece[:delta_size])
                self._data[0] = current_piece[delta_size:]
                self._data_length -= delta_size
                self._external_data_length.result -= delta_size

            # return memoryview(self._joiner.join(result_data))
            return self._joiner.join(result_data)

    def read_data(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long first_piece_len
        cdef object current_piece
        cdef object result_data
        cdef long long result_size
        cdef object another_piece
        cdef long long another_piece_len
        cdef long long delta_size
        cdef long long piece_index

        first_piece_len = len(self._data[0])
        if size < first_piece_len:
            current_piece = memoryview(self._data[0])
            return current_piece[:size]
        elif size == first_piece_len:
            return memoryview(self._data[0])
        else:
            result_data = deque()
            result_size = 0

            piece_index = 0
            while self._real_size and (result_size < size):
                another_piece = self._data[index]
                index += 1
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                result_size += another_piece_len
                if (result_size < size) and ((len(self._data[index]) + result_size) > size):
                    break

            if result_size < size:
                delta_size = size - result_size
                current_piece = memoryview(self._data[index])
                result_data.append(current_piece[:delta_size])

            # return memoryview(self._joiner.join(result_data))
            return self._joiner.join(result_data)

    # @profile
    def get_data_pieces(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long first_piece_len
        cdef object current_piece
        cdef object result_data
        cdef long long result_size
        cdef object another_piece
        cdef long long another_piece_len
        cdef long long delta_size

        first_piece_len = len(self._data[0])
        if size <= first_piece_len:
            self._data_length -= size
            self._external_data_length.result -= size
            self._real_size -= 1
            return self._data.popleft()
        else:
            result_data = deque()
            result_size = 0

            while self._real_size and (result_size < size):
                another_piece = self._data.popleft()
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                self._data_length -= another_piece_len
                self._external_data_length.result -= another_piece_len
                self._real_size -= 1
                result_size += another_piece_len

            return result_data

    def read_data_pieces(self, long long size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        cdef long long first_piece_len
        cdef object current_piece
        cdef object result_data
        cdef long long result_size
        cdef object another_piece
        cdef long long another_piece_len
        cdef long long delta_size
        cdef long long piece_index

        first_piece_len = len(self._data[0])
        if size <= first_piece_len:
            return memoryview(self._data[0])
        else:
            result_data = deque()
            result_size = 0

            piece_index = 0
            while self._real_size and (result_size < size):
                another_piece = memoryview(self._data[index])
                index += 1
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                result_size += another_piece_len

            return result_data

    def remove(self):
        if not self._removed:
            self._external_data_length.result -= self._data_length
            self._removed = True

    def __copy__(self):
        return DynamicListOfPiecesDequeWithLengthControl(
            self._init_param__joiner, self._init_param__external_data_length)

    def __del__(self):
        self.remove()


def get_num_pieces_with_full_size_lesser_than(object iter_object, long long full_size):
    cdef long long index = -1
    cdef long long result_size = 0
    cdef long long piece_len = 0
    cdef object piece

    if iter_object:
        result_size = 0
        index = 0
        for piece in iter_object:
            piece_len = len(piece)
            if (result_size + piece_len) > full_size:
                break
            else:
                result_size += piece_len
                index += 1
    return index


def get_num_pieces_with_full_size_lesser_than_with_offset(object iter_object, long long full_size,
                                                                              long long offset):
    cdef long long pieces_qnt = -1
    cdef long long object_size = 0
    cdef long long result_data_length = 0
    cdef long long piece_num = 0
    cdef object piece
    cdef long long piece_length = 0

    if iter_object:
        object_size = len(iter_object)
        result_data_length = 0
        pieces_qnt = 0
        for piece_num in range(offset, object_size):
            piece = iter_object[piece_num]
            piece_length = len(piece)
            if (result_data_length + piece_length) > full_size:
                break
            else:
                result_data_length += piece_length
                pieces_qnt += 1
    return pieces_qnt



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['get_uvicorn_awaitable']


from uvicorn.main import *
from h11._connection import DEFAULT_MAX_INCOMPLETE_EVENT_SIZE


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class UvicornStartupFailureError(Exception):
    pass


def get_uvicorn_awaitable(
    app: typing.Union["ASGIApplication", typing.Callable, str],
    *,
    host: str = "127.0.0.1",
    port: int = 8000,
    uds: typing.Optional[str] = None,
    fd: typing.Optional[int] = None,
    loop: LoopSetupType = "auto",
    http: typing.Union[typing.Type[asyncio.Protocol], HTTPProtocolType] = "auto",
    ws: typing.Union[typing.Type[asyncio.Protocol], WSProtocolType] = "auto",
    ws_max_size: int = 16777216,
    ws_ping_interval: typing.Optional[float] = 20.0,
    ws_ping_timeout: typing.Optional[float] = 20.0,
    ws_per_message_deflate: bool = True,
    lifespan: LifespanType = "auto",
    interface: InterfaceType = "auto",
    reload: bool = False,
    reload_dirs: typing.Optional[typing.Union[typing.List[str], str]] = None,
    reload_includes: typing.Optional[typing.Union[typing.List[str], str]] = None,
    reload_excludes: typing.Optional[typing.Union[typing.List[str], str]] = None,
    reload_delay: float = 0.25,
    workers: typing.Optional[int] = None,
    env_file: typing.Optional[typing.Union[str, os.PathLike]] = None,
    log_config: typing.Optional[
        typing.Union[typing.Dict[str, typing.Any], str]
    ] = LOGGING_CONFIG,
    log_level: typing.Optional[typing.Union[str, int]] = None,
    access_log: bool = True,
    proxy_headers: bool = True,
    server_header: bool = True,
    date_header: bool = True,
    forwarded_allow_ips: typing.Optional[typing.Union[typing.List[str], str]] = None,
    root_path: str = "",
    limit_concurrency: typing.Optional[int] = None,
    backlog: int = 2048,
    limit_max_requests: typing.Optional[int] = None,
    timeout_keep_alive: int = 5,
    ssl_keyfile: typing.Optional[str] = None,
    ssl_certfile: typing.Optional[typing.Union[str, os.PathLike]] = None,
    ssl_keyfile_password: typing.Optional[str] = None,
    ssl_version: int = SSL_PROTOCOL_VERSION,
    ssl_cert_reqs: int = ssl.CERT_NONE,
    ssl_ca_certs: typing.Optional[str] = None,
    ssl_ciphers: str = "TLSv1",
    headers: typing.Optional[typing.List[typing.Tuple[str, str]]] = None,
    use_colors: typing.Optional[bool] = None,
    app_dir: typing.Optional[str] = None,
    factory: bool = False,
    h11_max_incomplete_event_size: int = DEFAULT_MAX_INCOMPLETE_EVENT_SIZE,
) -> None:
    reload = False
    workers = None

    if app_dir is not None:
        sys.path.insert(0, app_dir)

    config = Config(
        app,
        host=host,
        port=port,
        uds=uds,
        fd=fd,
        loop=loop,
        http=http,
        ws=ws,
        ws_max_size=ws_max_size,
        ws_ping_interval=ws_ping_interval,
        ws_ping_timeout=ws_ping_timeout,
        ws_per_message_deflate=ws_per_message_deflate,
        lifespan=lifespan,
        interface=interface,
        reload=reload,
        reload_dirs=reload_dirs,
        reload_includes=reload_includes,
        reload_excludes=reload_excludes,
        reload_delay=reload_delay,
        workers=workers,
        env_file=env_file,
        log_config=log_config,
        log_level=log_level,
        access_log=access_log,
        proxy_headers=proxy_headers,
        server_header=server_header,
        date_header=date_header,
        forwarded_allow_ips=forwarded_allow_ips,
        root_path=root_path,
        limit_concurrency=limit_concurrency,
        backlog=backlog,
        limit_max_requests=limit_max_requests,
        timeout_keep_alive=timeout_keep_alive,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        ssl_keyfile_password=ssl_keyfile_password,
        ssl_version=ssl_version,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs,
        ssl_ciphers=ssl_ciphers,
        headers=headers,
        use_colors=use_colors,
        factory=factory,
        h11_max_incomplete_event_size=h11_max_incomplete_event_size,
    )
    server = Server(config=config)

    async def uvicorn_awaitable():
        await server.serve()

        if config.uds and os.path.exists(config.uds):
            os.remove(config.uds)  # pragma: py-win32

        if not server.started and not config.should_reload and config.workers == 1:
            raise UvicornStartupFailureError
    
    return uvicorn_awaitable()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from enum import Enum


class Fields(Enum):
    command_name = 0  # See Commands class. Optional: if ommited then: 1) if `in_response_to` present then default is `send_response`; 2) if `in_response_to` is not present then default is `send_request`
    request_id = 1  # id of either request or response (requests and responses must have an independent counters)
    in_response_to = 2  # current response is a response to request with this ID. Optional: can be present only in `send_response`
    is_response_required = 3  # int. 1 = required; 0 = not required. Optional: if ommited then default is 0 (not required). This gives us ability to send smallest response-less request as possible. Even is not required, response still can be send. If required - response must be send.
    data_serializer_id = 4  # If data was serialized explicitly with some serializer (Pickle for example). Optional.
    data = 5  # Data. Format depends on request type (command type). Optional: if ommited then default is None.


class Commands(Enum):
    # suggest_best_serializers_list = 0  # client suggests list of it's best serializers (first is fastest)
    # declare_best_serializers_list = 1  # server respondes with a sublist of apropriate sugested serializers (first is fastest)
    declare_client_node = 0
    declare_server_node = 1
    declare_service_class = 2  # client sends service class name, it's unique ID (int starting from 0, generated on the client side), it's module importable str (path) and it's module full file path
    declare_service_request_class = 3  # client sends service ID, service request class name and it's unique ID (int starting from 0, generated on the client side)
    send_request = 4  # raw request
    send_service_request = 5
    send_service_request_with_request_class = 6
    send_response = 6


# === Commands ====================
class CommandDataFieldsDeclareClientNode(Enum):
    platform_name = 0
    suggested_serializers = 1


class CommandDataFieldsDeclareServerNode(Enum):
    platform_name = 0
    declared_serializers = 1


class CommandDataFieldsDeclareServiceClass(Enum):
    local_id = 0
    class_name = 1
    module_importable_str = 2


class CommandDataFieldsDeclareServiceRequestClass(Enum):
    local_id = 0
    class_name = 1
    module_importable_str = 2
    properties_tuple = 3


class CommandDataFieldsRequest:
    args = 0
    kwargs = 1


class CommandDataFieldsServiceRequest:
    service_class_id = 0
    args = 1
    kwargs = 2


class CommandDataFieldsServiceRequestWithRequestClass:
    service_class_id = 0
    request_class_id = 1
    properties_tuple = 3



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class ValueExistence:
    __slots__ = ('existence', 'result')

    def __init__(self, existence: bool, result: Any):
        self.existence: bool = existence
        self.result = result

    def __bool__(self):
        return self.existence

    def __nonzero__(self):
        return self.__bool__()

    def __str__(self):
        return 'ValueExistence({}, {})'.format(self.existence, self.result)

    def __getstate__(self):
        return self.existence, self.result

    def __setstate__(self, state):
        existence, value = state
        self.value = value
        self.existence = existence
    
    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, ValueExistence):
            return (self.existence == __value.existence) and (self.value == __value.value)
        else:
            if self.existence:
                return self.value == __value
            else:
                return False
    
    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)


class ValueCache(ValueExistence):
    __slots__ = tuple()

    def __init__(self):
        super(ValueCache, self).__init__(False, None)

    def __call__(self, *args, **kwargs):
        self.existence = False

    def get(self):
        return self.result

    def set(self, new_result):
        self.existence = True
        self.result = new_result

    def __getstate__(self):
        return self.existence, self.result

    def __setstate__(self, state):
        existence, value = state
        self.value = value
        self.existence = existence
    
    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, ValueExistence):
            return (self.existence == __value.existence) and (self.value == __value.value)
        else:
            if self.existence:
                return self.value == __value
            else:
                return False
    
    def __ne__(self, __value: object) -> bool:
        return not self.__eq__(__value)


class ValueType:
    __slots__ = ('type_id', 'result')

    def __init__(self, type_id, result):
        self.type_id = type_id
        self.result = result

    def __eq__(self, other):
        # "__ne__() delegates to __eq__() and inverts the result"
        if isinstance(other, ValueType):
            return self.type_id == other.type_id
        else:
            return self.type_id == other

    def __getstate__(self):
        return self.type_id, self.result

    def __setstate__(self, state):
        self.type_id, self.result = state



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.performance_test_lib import PrecisePerformanceTestTracer
from cengal.help_tools import json_to_printable_string, make_readable_json
import json
from cengal.data_manipulation.serialization import test_data_factory, TestDataType
from cengal.data_manipulation.objects_counter import objects_counter, objects_counter_recursive, object_counter_uni
from cengal.introspection.inspect import intro_func_repr
from pprint import pprint


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def test_objects_counter():
    print(f'\n<<< {intro_func_repr()} >>>')
    data = {
            1: 'Hello',
            2: ('W', 0),
            3: [
                'r',
                1,
                {
                    'd': '.',
                    'dd': {
                        43: [0],
                        15: {
                            'world': 42
                        }
                    }
                }
            ],
            'To all!': '!!1'
        }
    print(make_readable_json(json.dumps(data)))
    
    num_of_objects = objects_counter(data)
    print('<<objects_counter>>: Number of objects in container: {}'.format(num_of_objects))
    assert(16 == num_of_objects)

    num_of_objects = objects_counter_recursive(data)
    print('<<objects_counter_recursive>>: Number of objects in container: {}'.format(num_of_objects))
    assert(16 == num_of_objects)

    num_of_objects = object_counter_uni(data)
    print('<<object_counter_uni>>: Number of objects in container: {}'.format(num_of_objects))
    assert(16 == num_of_objects)


def test_objects_counter_performance():
    print(f'\n<<< {intro_func_repr()} >>>')
    counter = objects_counter

    print()
    data = test_data_factory(TestDataType.small)
    # pprint(data)
    print('Small: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.deep_small)
    print('Small, Deep: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.large)
    # pprint(data)
    print('Large: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.deep_large)
    print('Large, Deep: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))


def test_objects_counter_recursive_performance():
    print(f'\n<<< {intro_func_repr()} >>>')
    counter = objects_counter_recursive

    print()
    data = test_data_factory(TestDataType.small)
    # pprint(data)
    print('Small: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.deep_small)
    print('Small, Deep: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.large)
    # pprint(data)
    print('Large: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    print()
    data = test_data_factory(TestDataType.deep_large)
    print('Large, Deep: {}'.format(counter(data)))
    tr = PrecisePerformanceTestTracer(1.0, turn_off_gc=True)
    while tr.iter():
        counter(data)
    with tr as fast_iter:
        for i in fast_iter:
            counter(data)
    print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))


def main():
    test_objects_counter()
    test_objects_counter_performance()
    test_objects_counter_recursive_performance()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import Interface, current_interface, CoroID
from nicegui.elements.mixins.text_element import TextElement
from nicegui.elements.select import Select
from nicegui.elements.tabs import Tab
from cengal.text_processing.text_translator import TranslationLanguageId, TextEntityId, TextTranslator, TranslationLanguageChooser, \
    TranslationLanguageMapper, TranslatableText, tt, TranslateMe, TMe, tme, TranslatableTextElement, TTE
from cengal.introspection.inspect import is_callable
from cengal.file_system.path_manager import path_relative_to_src
from typing import Optional, Tuple, Dict, Any, List, Union, Callable, Hashable, Set
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, put_coro
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoroRequest, CoroutineNotFoundError
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBusRequest
from cengal.text_processing.text_translator.versions.v_1.text_translator import TranslationLanguageChooser


class NiceguiTranslatableTextElement(TranslatableTextElement[TextElement]):
    def __init__(self, text_translation_language_chooser: TranslationLanguageChooser) -> None:
        super().__init__(text_translation_language_chooser)
    
    def __call__(self, text_element: TextElement) -> TextElement:
        if text_element is None:
            return None
        
        if isinstance(text_element, TextElement):
            if isinstance(text_element.text, TranslateMe):
                if text_element.text:
                    self.elements_and_their_translatable_text[text_element] = text_element.text
                
                text_element.text = text_element.text.to_str(self.text_translator, self.text_translation_language_chooser.end_lang)
        elif isinstance(text_element, Select):
            if 'label' in text_element._props:
                text: TranslateMe = text_element._props['label']
                if isinstance(text, TranslateMe):
                    if text:
                        self.elements_and_their_translatable_text[text_element] = text
                    
                    text_element._props['label'] = text.to_str(self.text_translator, self.text_translation_language_chooser.end_lang)
        elif isinstance(text_element, Tab):
            text: TranslateMe = text_element._props['label']
            if isinstance(text, TranslateMe):
                if text:
                    self.elements_and_their_translatable_text[text_element] = text
                
                text_element._props['label'] = text.to_str(self.text_translator, self.text_translation_language_chooser.end_lang)
        
        return text_element

    def set_text(self, text_element: TextElement, text: Union[TranslateMe, str]) -> TextElement:
        if text_element is None:
            return None
        
        need_to_remove = False
        if isinstance(text, TranslateMe):
            if text:
                self.elements_and_their_translatable_text[text_element] = text
                text_element.text = text.to_str(self.text_translator, self.text_translation_language_chooser.end_lang)
            else:
                need_to_remove = True
        else:
            need_to_remove = True
        
        if need_to_remove:
            del self.elements_and_their_translatable_text[text_element]
            text_element.text = text
        
        return text_element


NTTE = NiceguiTranslatableTextElement


def setup_translation(text_dictionary_path: str):
    # with open(path_relative_to_src('./text_dictionary.json'), 'rb') as f:
    with open(text_dictionary_path, 'rb') as f:
        text_dictionary_data = f.read()

    text_translator: TextTranslator = TextTranslator.from_json(text_dictionary_data)
    translation_language_mapper: TranslationLanguageMapper = TranslationLanguageMapper(text_translator.decoded_data['translation_language_map'],
                                                                                    text_translator.decoded_data['default_language'])
    return text_translator, translation_language_mapper


def create_translatable_text_element(text_translator, translation_language_mapper):
    translation_language_chooser: TranslationLanguageChooser = TranslationLanguageChooser(text_translator, translation_language_mapper)
    # translation_language_chooser.lang = text_translator.decoded_data['original_text_language']
    return NiceguiTranslatableTextElement(translation_language_chooser)


async def acreate_translatable_text_element(text_translator, translation_language_mapper):
    translation_language_chooser: TranslationLanguageChooser = TranslationLanguageChooser(text_translator, translation_language_mapper)
    # translation_language_chooser.lang = text_translator.decoded_data['original_text_language']

    translatable_text_element = NiceguiTranslatableTextElement(translation_language_chooser)
    await translatable_text_element.aregister_on_lang_changed_handler()
    return translatable_text_element



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from cengal.time_management.relative_time.approximate_representation.versions.v_0 import ApproximateTimeRepresentation, FULL_FORMATTER

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class TestApproximateTimeRepresentation(unittest.TestCase):
    def setUp(self):
        self.tr = ApproximateTimeRepresentation(3.045678, crop=False)
        self.trc = ApproximateTimeRepresentation(3.045678, crop=True)

    # def test_all_parameters_are_present(self):
    #     tr = ApproximateTimeRepresentation(3.045678, crop=True)
    #     tr_res = tr.format(FULL_FORMATTER)
    #     print(tr_res)
    #     tr_res = tr.format_cropped(FULL_FORMATTER)
    #     print(tr_res)
    #     tr_res = str(tr)
    #     print(tr_res)
    #     expected_result = '|(years)%4%; - %2%; - %1%|.'
    #     self.assertEqual(1, 1, 'incorrect result')

    def test__uncroped__format(self):
        tr_res = self.tr.format()
        expected_result = 'Y0-0-0 0:0:3.045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__uncroped__format__full(self):
        tr_res = self.tr.format(FULL_FORMATTER)
        expected_result = 'Millennia: 0, Centuries: 0, Decades: 0, Years: 0, Months: 0, Weeks: 0, Days: 0, Hours: 0, Minutes: 0, Seconds: 3, Milliseconds: 000, Microseconds: 045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__uncroped__format_cropped(self):
        tr_res = self.tr.format_cropped()
        expected_result = 'Y0-0-0 0:0:3.045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__uncroped__format_cropped__full(self):
        tr_res = self.tr.format_cropped(FULL_FORMATTER)
        expected_result = 'Years: 0, Months: 0, Days: 0, Hours: 0, Minutes: 0, Seconds: 3, Microseconds: 045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__uncroped__str(self):
        tr_res = str(self.tr)
        expected_result = 'ApproximateTimeRepresentation{Millennia: 0, Centuries: 0, Decades: 0, Years: 0, Months: 0, Weeks: 0, Days: 0, Hours: 0, Minutes: 0, Seconds: 3, Milliseconds: 000, Microseconds: 045678}'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__croped__format(self):
        tr_res = self.trc.format()
        expected_result = 'Y0-0-0 0:0:3.045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__croped__format__full(self):
        tr_res = self.trc.format(FULL_FORMATTER)
        expected_result = 'Millennia: 0, Centuries: 0, Decades: 0, Years: 0, Months: 0, Weeks: 0, Days: 0, Hours: 0, Minutes: 0, Seconds: 3, Milliseconds: 000, Microseconds: 045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__croped__format_cropped(self):
        tr_res = self.trc.format_cropped()
        expected_result = 's3.045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__croped__format_cropped__full(self):
        tr_res = self.trc.format_cropped(FULL_FORMATTER)
        expected_result = 'Seconds: 3, Microseconds: 045678'
        self.assertEqual(tr_res, expected_result, 'incorrect result')

    def test__croped__str(self):
        tr_res = str(self.trc)
        expected_result = 'ApproximateTimeRepresentation{Millennia: 0, Centuries: 0, Decades: 0, Years: 0, Months: 0, Weeks: 0, Days: 0, Hours: 0, Minutes: 0, Seconds: 3, Milliseconds: 000, Microseconds: 045678}'
        self.assertEqual(tr_res, expected_result, 'incorrect result')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['LazyPrint', 'lazy_print', 'lprint', 'lp']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


from time import perf_counter
from typing import Dict, List, Tuple

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import TimerFuncRunner, add_timer_func_run_from_other_service, discard_timer_func_run_from_other_service

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class LazyPrint(TypedService[None]):
    def __init__(self, loop):
        super().__init__(loop)
        self.requests = list()
        self.update_period = 1 / 60
        self.update_timer_request = None

    def single_task_registration_or_immediate_processing(self, *args, **kwargs) -> Tuple[(bool, None, None)]:
        self.requests.append((args, kwargs))
        self.try_add_timer()
        return (True, None, None)

    def full_processing_iteration(self):
        requests_buff = self.requests
        self.requests = type(self.requests)()
        for args, kwargs in requests_buff:
            kwargs['flush'] = False
            print(*args, **kwargs)

        print(end='', flush=True)
        self.add_timer()
        self.make_dead()

    def timer_handler(self):
        self.update_timer_request = None
        if self.in_work_impl():
            self.make_live()
    
    def add_timer(self):
        self.update_timer_request = add_timer_func_run_from_other_service(self, True, self.update_period, self.timer_handler)
    
    def try_add_timer(self):
        if self.update_timer_request is None: self.add_timer()

    def _put_direct_request(self, *args, **kwargs):
        self.requests.append((args, kwargs))
        self.try_add_timer()

    def in_work_impl(self) -> bool:
        return bool(self.requests)

    def in_work(self) -> bool:
        result: bool = self.in_work_impl()
        return self.thrifty_in_work(result)
    
    def destroy(self):
        try:
            discard_timer_func_run_from_other_service(self, self.update_timer_request)
        except CoroSchedulerIsCurrentlyDestroingError:
            pass


def lazy_print(*args, **kwargs):
    fallback = False
    try:
        cs = current_coro_scheduler()
        if cs._destroyed:
            fallback = True
    except OutsideCoroSchedulerContext:
        fallback = True

    if fallback:
        fallback = False
        try:
            cs: CoroSchedulerType = primary_coro_scheduler()
            if cs._destroyed:
                fallback = True
        except PrimaryCoroSchedulerWasNotSet:
            fallback = True

    if fallback:
        fallback = False
        print(*args, **kwargs)
    else:
        cs.get_service_instance(LazyPrint)._put_direct_request(*args, **kwargs)


lprint = lazy_print
lp = lazy_print



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import socket
import errno
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
from cengal.base.classes import BaseClassSettings



__author__ = 'ButenkoMS <gtalk@butenkoms.space>'


SET_OF_CONNECTION_ERRORS = {errno.ECONNRESET, errno.ECONNREFUSED, errno.ECONNABORTED, errno.EPIPE, errno.ESHUTDOWN}
INET_TYPE_CONNECTIONS = {socket.AF_INET, socket.AF_INET6}


try:
    BlockingIOError
except NameError:
    class BlockingIOError(OSError):
        pass

try:
    InterruptedError
except NameError:
    class InterruptedError(OSError):
        pass

try:
    ConnectionError
except NameError:
    class ConnectionError(OSError):
        pass

try:
    BrokenPipeError
except NameError:
    class BrokenPipeError(ConnectionError):
        pass

try:
    ConnectionAbortedError
except NameError:
    class ConnectionAbortedError(ConnectionError):
        pass

try:
    ConnectionRefusedError
except NameError:
    class ConnectionRefusedError(ConnectionError):
        pass

try:
    ConnectionResetError
except NameError:
    class ConnectionResetError(ConnectionError):
        pass


MESSAGE_SIZE_LEN = 8
SERVER_ANSWER__KEYWORD_ACCEPTED = b'OK'


class SimpleNetworkError(Exception):
    pass


class ConnectionDirectionRole:
    server = 0
    client = 1


# class IPVersion:
#     ipv4 = 0
#     ipv6 = 1


# class ConnectionSettings(BaseClassSettings):
#     def __init__(self, host=None, port=None, keyword=None, direction_role=None, ip_version=None):
#         '''
#         :param host: 'localhost', '192.168.0.1', ect.
#         :param port: 8080
#         :param keyword: b'sdlkfj s894 saf 84ewksdhf sdf'
#         :param direction_role: ConnectionDirectionRole()
#         :param ip_version: IPVersion()
#         :return:
#         '''
#         self.host = host or str()
#         self.port = port or None
#         self.keyword = keyword or bytes()
#         self.direction_role = direction_role or None
#         self.ip_version = ip_version or None
#
#     def check(self):
#         if (self.port is None) or (self.direction_role is None) or (self.ip_version is None):
#             raise Exception('port can\'t be None; direction_role can\'t be None; ip_version can\'t be None.')


class ConnectionSettings(BaseClassSettings):
    def __init__(self, direction_role: ConnectionDirectionRole=None, socket_address=None, keyword: bytes=None,
                 socket_family=socket.AF_INET, socket_type=socket.SOCK_STREAM, socket_protocol=0, socket_fileno=None):
        '''
        :param direction_role: ConnectionDirectionRole()
        :param socket_address: './main.server.AF_UNIX.socket', ('localhost', 8080), ('::', 50007, 0, 0), , ect.
        :param keyword: b'sdlkfj s894 saf 84ewksdhf sdf'. Can be None for a Super Server
        :param socket_family: AF_INET (the default), AF_INET6, AF_UNIX, AF_CAN or AF_RDS
        :param socket_type: SOCK_STREAM (the default), SOCK_DGRAM, SOCK_RAW or perhaps one of the other SOCK_ constants
        :param socket_protocol: in the case where the address family is AF_CAN the protocol should be one of CAN_RAW or
            CAN_BCM
        :param socket_fileno: If fileno is specified, the other arguments are ignored, causing the socket with the
            specified file descriptor to return
        '''
        self.direction_role = direction_role
        self.keyword = keyword
        self.socket_address = socket_address
        self.socket_family = socket_family
        self.socket_type = socket_type
        self.socket_protocol = socket_protocol
        self.socket_fileno = socket_fileno
        self.expected_clients_with_empty_output_fifo = set()


class IOCoreMemoryManagement:
    def __init__(self):
        self.global__data_size_limit = ResultExistence(True, 2 * 1024**3)

        self.global_in__data_size_limit = ResultExistence(True, 512 * 1024**2)
        self.global_in__data_full_size = ResultExistence(True, 0)
        self.global_in__deletable_data_full_size = ResultExistence(True, 0)

        self.global_out__data_size_limit = ResultExistence(True, 512 * 1024**2)
        self.global_out__data_full_size = ResultExistence(True, 0)
        self.global_out__deletable_data_full_size = ResultExistence(True, 0)

    def link_to(self, parent):
        self.global__data_size_limit = parent.global__data_size_limit

        self.global_in__data_size_limit = parent.global_in__data_size_limit
        self.global_in__data_full_size = parent.global_in__data_full_size
        self.global_in__deletable_data_full_size = parent.global_in__deletable_data_full_size

        self.global_out__data_size_limit = parent.global_out__data_size_limit
        self.global_out__data_full_size = parent.global_out__data_full_size
        self.global_out__deletable_data_full_size = parent.global_out__deletable_data_full_size



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['WaitCoro', 'PutSingleCoroParams', 'PSCP', 'WaitCoroRequest', 'CoroutineNotFoundError', 'SubCoroutineNotFoundError', 'TimeoutError', 'SubTimeoutError']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro_list import PutSingleCoroParams, PSCP
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import timer_func_run_on
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import kill_coro_on
from cengal.introspection.inspect import get_exception
from cengal.data_manipulation.conversion.reinterpret_cast import reinterpret_cast
from typing import Any, Optional, Sequence, Tuple, Dict, Set, Union, List, overload, Type


class CoroutineNotFoundError(Exception):
    pass


class SubCoroutineNotFoundError(CoroutineNotFoundError):
    pass


class TimeoutError(Exception):
    pass


class SubTimeoutError(TimeoutError):
    pass


class ServParams:
    @overload
    def __init__(self, service_request_type: Type[TypedServiceRequest[ServiceResponseTypeVar]], *args, **kwargs) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_request: TypedServiceRequest[ServiceResponseTypeVar]) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_type: Type[TypedService[ServiceResponseTypeVar]], *args, **kwargs) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_type: ServiceType, service_request: TypedServiceRequest[ServiceResponseTypeVar]) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_request_type: Type[ServiceRequest], *args, **kwargs) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_request: ServiceRequest) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_type: ServiceType, *args, **kwargs) -> ServiceResponseTypeVar: ...

    @overload
    def __init__(self, service_type: ServiceType, service_request: ServiceRequest) -> ServiceResponseTypeVar: ...

    def __init__(self, service_type, *args, **kwargs) -> None:
        self.service_type: NormalizableServiceType = service_type
        self.args: Tuple = args
        self.kwargs: Dict = kwargs
    
    def __call__(self) -> Tuple[NormalizableServiceType, Tuple, Dict]:
        return self.service_type, self.args, self.kwargs


SP = ServParams


class WaitCoroRequest(ServiceRequest):
    def __init__(self, timeout: Optional[float] = None, kill_on_timeout: bool = True, tree: bool = True, result_required: bool = True):
        super().__init__()
        self.provide_to_request_handler = True
        self.timeout: Optional[float] = timeout
        self.kill_on_timeout: bool = kill_on_timeout
        self.tree: bool = tree
        self.result_required: bool = result_required

    def single(self, coro_id: CoroID) -> Union[Any, None]:
        return self._save(0, coro_id)

    def list(self, coro_list: Sequence[CoroID]) -> ServiceRequest:
        return self._save(1, coro_list)

    def atomic(self, coro_list: Sequence[CoroID]) -> ServiceRequest:
        return self._save(2, coro_list)

    def fastest(self, coro_list: Sequence[CoroID], num: int = 1, measure_time: bool = False) -> ServiceRequest:
        return self._save(3, coro_list, num, measure_time)

    def put_single(self, coro_worker: Worker, *args, **kwargs) -> ServiceRequest:
        return self._save(4, coro_worker, *args, **kwargs)

    def put_list(self, coro_list: Sequence[PutSingleCoroParams]) -> List[Tuple[CoroID, Any, Optional[Exception]]]:
        return self._save(5, coro_list)

    def put_atomic(self, coro_list: Sequence[PutSingleCoroParams]) -> ServiceRequest:
        return self._save(6, coro_list)

    def put_fastest(self, coro_list: Sequence[PutSingleCoroParams], num: int=1, measure_time: bool=False) -> ServiceRequest:
        return self._save(7, coro_list, num, measure_time)

    def serv_list(self, serv_params_list: Sequence[ServParams]) -> List[Tuple[CoroID, Any, Optional[Exception]]]:
        """Creates a coroutine for each service request and waits for the result of each of them.

        Args:
            serv_params_list (Sequence[ServParams]): _description_

        Returns:
            List[Tuple[CoroID, Any, Optional[Exception]]]: _description_
        """        
        return self._save(8, serv_params_list)

    def serv_atomic(self, serv_params_list: Sequence[ServParams]) -> ServiceRequest:
        """Creates a coroutine for each service request and waits for the result of each of them. If one of the coroutines fails, the others are killed which may lead to cancel request processing in some services (see documentation of the service you are trying to use here).

        Args:
            serv_params_list (Sequence[ServParams]): _description_

        Returns:
            ServiceRequest: _description_
        """        
        return self._save(9, serv_params_list)

    def serv_fastest(self, serv_params_list: Sequence[ServParams], num: int=1, measure_time: bool=False) -> ServiceRequest:
        """Creates a coroutine for each service request and waits for the result of each of them. When one of coroutines finished, the others are killed which may lead to cancel request processing in some services (see documentation of the service you are trying to use here)

        Args:
            serv_params_list (Sequence[ServParams]): _description_
            num (int, optional): _description_. Defaults to 1.
            measure_time (bool, optional): _description_. Defaults to False.

        Returns:
            ServiceRequest: _description_
        """        
        return self._save(10, serv_params_list, num, measure_time)

    def serv_and_forget_single(self, serv_params: ServParams) -> None:
        """Creates a coroutine for a service request and returns immediately.

        Args:
            serv_params (ServParams): _description_

        Returns:
            _type_: _description_
        """        
        return self._save(11, serv_params)

    def serv_and_forget_list(self, serv_params_list: Sequence[ServParams]) -> None:
        """Creates a coroutine for each service request and returns immediately.

        Args:
            serv_params_list (Sequence[ServParams]): _description_

        Returns:
            _type_: _description_
        """        
        return self._save(12, serv_params_list)


class SingleMethod(ServiceRequestMethodMixin):

    def __init__(self, service):
        super().__init__(service)
        self.single_called_by: Dict[CoroID, CoroID] = dict()  # Dict[CoroID, CoroID] # key - callable; value - requester
        self.new_single_results: Set[Tuple[CoroID, Any, Optional[BaseException]]] = set()  # (id, result, exception)
        self.result_required_by: Dict[CoroID, bool] = dict()

    def __call__(self, request: WaitCoroRequest, coro_id: CoroID) -> ServiceProcessingResponse:
        requester_id: CoroID = self.service.current_caller_coro_info.coro_id
        coro: CoroWrapperBase = self.service._loop.get_coro(coro_id)
        if coro is None:
            return (True, None, CoroutineNotFoundError(coro_id))

        coro.add_on_coro_del_handler(self._on_coro_del_handler)
        self.single_called_by[coro_id] = requester_id
        self.result_required_by[requester_id] = request.result_required
        timeout: Optional[float] = request.timeout
        if timeout is not None:
            def timeout_handler(coro_id: CoroID, kill_on_timeout: bool, tree: bool):
                if coro_id in self.single_called_by:
                    self.new_single_results.add((coro_id, None, TimeoutError(coro_id)))
                    self.service.make_live()
                    if kill_on_timeout:
                        kill_coro_on(get_interface_and_loop_with_explicit_loop(self.service._loop), coro_id, tree)
            
            timer_func_run_on(get_interface_and_loop_with_explicit_loop(self.service._loop), timeout, timeout_handler, coro_id, request.kill_on_timeout, request.tree)

        return (False, None, None)

    def full_processing_iteration(self):
        for coro_id, result, exception in self.new_single_results:
            try:
                requester_id: CoroID = self.single_called_by[coro_id]
                if self.result_required_by[requester_id]:
                    if CoroutineNotFoundError == type(exception):
                        reinterpret_cast(SubCoroutineNotFoundError, exception)
                    elif TimeoutError == type(exception):
                        reinterpret_cast(SubTimeoutError, exception)
                    
                    self.service.register_response(requester_id, result, exception)
                else:
                    self.service.register_response(requester_id, None, None)
                
                del self.result_required_by[requester_id]
                del self.single_called_by[coro_id]
            except KeyError:
                pass

        self.new_single_results = type(self.new_single_results)()

    def in_work(self) -> bool:
        return bool(self.new_single_results)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        if coro.coro_id in self.single_called_by:
            self.new_single_results.add((coro.coro_id, coro.last_result, coro.exception))
            self.service.make_live()
        
        return True


class ListMethod(ServiceRequestMethodMixin):

    def __init__(self, service):
        super().__init__(service)

    def __call__(self, request: WaitCoroRequest, coro_list: Sequence[Tuple[(Optional[CoroType], Worker, Tuple, Dict)]]) -> ServiceProcessingResponse:
        requester_id = self.service.current_caller_coro_info.coro_id
        for coro_id in coro_list:
            coro = self.service._loop.get_coro(coro_id)
            if coro is None:
                return (True, None, CoroutineNotFoundError(coro_id))
            else:
                coro.add_on_coro_del_handler(self._on_coro_del_handler)
                self.list_called_by[coro_id] = requester_id
                if requester_id not in self.list_wait_by_caller:
                    self.list_wait_by_caller[requester_id] = set()
                
                self.list_wait_by_caller[requester_id].add(coro_id)
                # timeout: Optional[float] = request.timeout
                # if timeout is not None:
                #     def timeout_handler(requester_id: CoroID, coro_id: CoroID, kill_on_timeout: bool, tree: bool):
                #         if coro_id in self.single_called_by:
                #             self.new_single_results.add((coro_id, None, TimeoutError(coro_id)))
                #             self.service.make_live()
                #             if kill_on_timeout:
                #                 kill_coro_on(get_interface_and_loop_with_explicit_loop(self.service._loop), coro_id, tree)
                    
                #     timer_func_run_on(get_interface_and_loop_with_explicit_loop(self.service._loop), timeout, timeout_handler, requester_id, coro_id, request.kill_on_timeout, request.tree)


        self.service.make_dead()
        return (False, None, None)

    def full_processing_iteration(self):
        raise NotImplementedError

    def in_work(self) -> bool:
        raise NotImplementedError

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        if coro.coro_id in self.single_called_by:
            self.new_list_results.add((coro.coro_id, coro.last_result, coro.exception))
            self.service.make_live()
        
        return True


class PutListMethod(ServiceRequestMethodMixin):

    def __init__(self, service):
        super().__init__(service)
        self.called_by = dict()
        self.caller_waiting_set: Dict[CoroID, Set[CoroID]] = dict()
        self.coro_indexes = dict()
        self.caller_results = dict()
        self.ready_requesters = set()

    def __call__(self, request: WaitCoroRequest, coro_list: Sequence[PutSingleCoroParams]) -> ServiceProcessingResponse:
        coroutines_list = list()
        results = list()
        requester_id = self.service.current_caller_coro_info.coro_id
        try:
            put_coro: PutCoro = self.service._loop.get_service_instance(PutCoro)
            for coro_request in coro_list:
                exception = None
                result_coro_id = None
                try:
                    coro_worker, args, kwargs = coro_request()
                    coro: CoroWrapperBase = put_coro.put_from_other_service(requester_id, coro_worker, *args, **kwargs)
                    coroutines_list.append(coro)
                    result_coro_id = coro.coro_id
                except:
                    exception = get_exception()

                results.append((result_coro_id, exception))
        except:
            return True, results, get_exception()
        else:
            if not coroutines_list:
                return True, results, None
            
            if requester_id not in self.caller_waiting_set:
                self.caller_waiting_set[requester_id] = set()
            if requester_id not in self.coro_indexes:
                self.coro_indexes[requester_id] = dict()
            if requester_id not in self.caller_results:
                self.caller_results[requester_id] = [
                 None] * len(coroutines_list)
            for index, coro in enumerate(coroutines_list):
                coro_id = coro.coro_id
                self.called_by[coro_id] = requester_id
                self.caller_waiting_set[requester_id].add(coro_id)
                self.coro_indexes[requester_id][coro_id] = index
                coro.add_on_coro_del_handler(self._on_coro_del_handler)

            timeout: Optional[float] = request.timeout
            if timeout is not None:
                def timeout_handler(requester_id: CoroID, kill_on_timeout: bool, tree: bool):
                    if requester_id not in self.caller_waiting_set:
                        return
                    
                    caller_waiting_set = self.caller_waiting_set[requester_id]
                    del self.caller_waiting_set[requester_id]
                    for coro_id in caller_waiting_set:
                        del self.called_by[coro_id]
                        index = self.coro_indexes[requester_id][coro_id]
                        del self.coro_indexes[requester_id][coro_id]
                        self.caller_results[requester_id][index] = (coro_id, None, coro.exception)  # TODO: should return TimeoutError exception instead
                        if kill_on_timeout:
                            kill_coro_on(get_interface_and_loop_with_explicit_loop(self.service._loop), coro_id, tree)

                    del self.coro_indexes[requester_id]
                    self.ready_requesters.add(requester_id)
                    self.service.make_live()
                
                timer_func_run_on(get_interface_and_loop_with_explicit_loop(self.service._loop), timeout, timeout_handler, requester_id, request.kill_on_timeout, request.tree)

            return False, None, None

    def full_processing_iteration(self):
        ready_requesters_buff = self.ready_requesters
        self.ready_requesters = type(ready_requesters_buff)()
        for requester_id in ready_requesters_buff:
            self.service.register_response(requester_id, self.caller_results[requester_id], None)
            del self.caller_results[requester_id]

    def in_work(self) -> bool:
        return bool(self.ready_requesters)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        coro_id = coro.coro_id
        if coro_id in self.called_by:
            requester_id = self.called_by[coro_id]
            del self.called_by[coro_id]
            self.caller_waiting_set[requester_id].remove(coro_id)
            if not self.caller_waiting_set[requester_id]:
                del self.caller_waiting_set[requester_id]
            
            index = self.coro_indexes[requester_id][coro_id]
            del self.coro_indexes[requester_id][coro_id]
            self.caller_results[requester_id][index] = (coro.coro_id, coro.last_result, coro.exception)
            if not self.coro_indexes[requester_id]:
                del self.coro_indexes[requester_id]
                self.ready_requesters.add(requester_id)
                self.service.make_live()

        return True


class WaitCoro(Service):

    def __init__(self, loop):
        super(WaitCoro, self).__init__(loop)
        self._single = SingleMethod(self)
        self._list = ListMethod(self)
        self._put_list = PutListMethod(self)

        self._request_workers = {
            0:self._single,
            1:self.not_implemented,
            2:self.not_implemented,
            3:self.not_implemented,
            4:self.not_implemented,
            5:self._put_list,
            6:self.not_implemented,
            7:self.not_implemented
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[WaitCoroRequest]) -> ServiceProcessingResponse:
        if request is not None:
            return self.resolve_request(request)
            
        return (True, None, WrongServiceRequestError())

    def full_processing_iteration(self):
        self._single.full_processing_iteration()
        self._put_list.full_processing_iteration()
        if not self.in_work():
            self.make_dead()

    def in_work(self) -> bool:
        result: bool = self._single.in_work() or self._put_list.in_work()
        return self.thrifty_in_work(result)

    def not_implemented(self):
        raise NotImplementedError


WaitCoroRequest.default_service_type = WaitCoro



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = [
    'TaskIsNotFinished', 'TaskIsFinished', 'Task', 
    'PutCoro', 'PutCoroRequest', 'put_current_from_other_service', 'put_from_other_service', 'put_root_from_other_service', 
    'put_coro_to', 'try_put_coro_to', 'aput_coro_to', 'atry_put_coro_to', 'put_coro', 'try_put_coro', 'aput_coro', 
    'atry_put_coro', 'travers_through_all_coro_children_on', 'try_travers_through_all_coro_children_on', 
    'travers_through_all_coro_children', 'try_travers_through_all_coro_children', 'get_set_of_all_coro_children_on', 
    'try_get_set_of_all_coro_children_on', 'get_set_of_all_coro_children', 'try_get_set_of_all_coro_children', 
    'travers_through_all_coro_parents_on', 'try_travers_through_all_coro_parents_on', 'travers_through_all_coro_parents', 
    'try_travers_through_all_coro_parents', 'get_set_of_all_coro_parents_on', 'try_get_set_of_all_coro_parents_on', 
    'get_set_of_all_coro_parents', 'try_get_set_of_all_coro_parents', 'get_set_of_all_coros_on', 
    'try_get_set_of_all_coros_on', 'get_set_of_all_coros', 'try_get_set_of_all_coros']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.introspection.inspect import get_exception, get_exception_tripple
from cengal.data_manipulation.tree_traversal import KeyMultiValueTreeTraversal, KeyValueTreeTraversal
from typing import Hashable, Tuple, List, Optional, Any, Union, cast, Dict, Callable, Set


class TaskIsNotFinished(Exception):
    pass


class TaskIsFinished(Exception):
    pass


class Task:
    def __init__(self, coro_wrapper: CoroWrapperBase):
        self._coro_wrapper = coro_wrapper
        self._coro_id = coro_wrapper.coro_id
        self._coro_wrapper.add_on_coro_del_handler(self._on_coro_del_handler)
        self._done: bool = False
        self._result: Any = None
        self._exception: Exception = None
        self._on_done_handlers: Set[Callable[['Task'], None]] = set()
    
    def _on_coro_del_handler(self, coro_wrapper: CoroWrapperBase) -> bool:
        self._result = self._coro_wrapper.last_result
        self._exception = self._coro_wrapper.exception
        self._coro_wrapper = None
        self._done = True
        for handler in self._on_done_handlers:
            handler(self)
        
        return True
    
    @property
    def coro_id(self):
        return self._coro_id
    
    @property
    def coro_wrapper(self):
        if self._done:
            raise TaskIsFinished()

        return self._coro_wrapper
    
    @property
    def result(self):
        if not self._done:
            raise TaskIsNotFinished()
        
        if self._exception is not None:
            raise self._exception
        
        return self._result
    
    def done(self) -> bool:
        return self._done
    
    def __bool__(self) -> bool:
        return self._done

    def __nonzero__(self):
        return self.__bool__()
    
    def add_on_done_handler(self, handler: Callable[['Task'], None]) -> bool:
        if self._done:
            handler(self)
        else:
            self._on_done_handlers.add(handler)
        
        return True


class PutCoroRequest(ServiceRequest):
    def turn_on_tree_monitoring(self, turn_on: bool) -> bool:
        return self._save(0, turn_on)

    def tree_monitoring_state(self) -> bool:
        return self._save(1)

    # TODO: An implementation required:
    def set_on_children_start_handler(self, parent_coro_id: CoroID, handler: Callable) -> ServiceRequest:
        return self._save(2, parent_coro_id, handler)

    # TODO: An implementation required:
    def set_on_children_del_handler(self, parent_coro_id: CoroID, handler: Callable) -> ServiceRequest:
        return self._save(3, parent_coro_id, handler)

    def put_background_coro(self, coro_worker: AnyWorker, *args, **kwargs) -> ServiceRequest:
        return self._save(4, coro_worker, args, kwargs)

    def task(self, coro_worker: AnyWorker, *args, **kwargs) -> ServiceRequest:
        return self._save(5, coro_worker, args, kwargs, False)

    def background_task(self, coro_worker: AnyWorker, *args, **kwargs) -> ServiceRequest:
        return self._save(5, coro_worker, args, kwargs, True)


class PutCoro(ServiceWithADirectRequestMixin, DualImmediateProcessingServiceMixin, TypedService[CoroID]):
    def __init__(self, loop: CoroSchedulerType):
        super(PutCoro, self).__init__(loop)
        self._tree_monitoring_turned_on: bool = None
        self._single_task_registration_or_immediate_processing_single_impl: Callable = None
        self._single_task_registration_or_immediate_processing_single_impl_impl: Callable = None
        self._turn_on_tree_monitoring(False)
        self._tree_children_by_parent: Dict[CoroID, Set[CoroID]] = dict()
        self._tree_parent_by_child: Dict[CoroID, CoroID] = dict()
        self._dead_coroutines: Set[CoroID] = set()

        self._request_workers = {
            0: self._turn_on_tree_monitoring,
            1: self._tree_monitoring_state,
            4: self._put_background_coro,
            5: self._task,
        }

        self.direct_requests: List[Tuple] = list()
    
    def single_task_registration_or_immediate_processing_single(
            self, coro_worker: AnyWorker, *args, **kwargs
    ) -> Tuple[bool, Optional[CoroID], Any]:
        try:
            coro = self._single_task_registration_or_immediate_processing_single_impl(coro_worker, *args, **kwargs)
        except:
            return True, None, get_exception()

        return True, coro.coro_id, None
    
    def _single_task_registration_or_immediate_processing_single_impl_with_tree(
            self, coro_worker: AnyWorker, *args, **kwargs
    ) -> CoroWrapperBase:
        return self._single_task_registration_or_immediate_processing_single_impl_with_tree_impl(self.current_caller_coro_info.coro_id, coro_worker, *args, **kwargs)
    
    def _single_task_registration_or_immediate_processing_single_impl_with_tree_impl(
            self, parent_coro_id: CoroID, coro_worker: AnyWorker, *args, **kwargs
    ) -> CoroWrapperBase:
        coro: CoroWrapperBase = self._loop.put_coro(coro_worker, *args, **kwargs)
        child_coro_id: CoroID = coro.coro_id
        if parent_coro_id not in self._tree_children_by_parent:
            self._tree_children_by_parent[parent_coro_id] = set()
        
        self._tree_children_by_parent[parent_coro_id].add(child_coro_id)
        self._tree_parent_by_child[child_coro_id] = parent_coro_id
        coro.add_on_coro_del_handler(self._on_coro_del_handler)
        return coro
    
    def _single_task_registration_or_immediate_processing_single_impl_without_tree(
            self, coro_worker: AnyWorker, *args, **kwargs
    ) -> CoroWrapperBase:
        return self._single_task_registration_or_immediate_processing_single_impl_without_tree_impl(self.current_caller_coro_info.coro_id, coro_worker, *args, **kwargs)
    
    def _single_task_registration_or_immediate_processing_single_impl_without_tree_impl(
            self, parent_coro_id: CoroID, coro_worker: AnyWorker, *args, **kwargs
    ) -> CoroWrapperBase:
        return self._loop.put_coro(coro_worker, *args, **kwargs)
    
    def _turn_on_tree_monitoring(self, turn_on: bool):
        previous_state: bool = self._tree_monitoring_turned_on
        self._tree_monitoring_turned_on = turn_on
        if turn_on:
            self._single_task_registration_or_immediate_processing_single_impl = self._single_task_registration_or_immediate_processing_single_impl_with_tree
            self._single_task_registration_or_immediate_processing_single_impl_impl = self._single_task_registration_or_immediate_processing_single_impl_with_tree_impl
        else:
            self._single_task_registration_or_immediate_processing_single_impl = self._single_task_registration_or_immediate_processing_single_impl_without_tree
            self._single_task_registration_or_immediate_processing_single_impl_impl = self._single_task_registration_or_immediate_processing_single_impl_without_tree_impl
        
        return True, previous_state, None
    
    def put_from_other_service(
            self, coro_id: CoroID, coro_worker: AnyWorker, *args, **kwargs
        ) -> CoroWrapperBase:
        return self._single_task_registration_or_immediate_processing_single_impl_impl(coro_id, coro_worker, *args, **kwargs)
    
    def put_task_from_other_service(
            self, coro_id: CoroID, coro_worker: AnyWorker, args, kwargs
        ) -> Task:
        coro: CoroWrapperBase = self._single_task_registration_or_immediate_processing_single_impl_impl(coro_id, coro_worker, *args, **kwargs)
        return Task(coro)
    
    def put_root_from_other_service(
            self, coro_worker: AnyWorker, *args, **kwargs
        ) -> CoroWrapperBase:
        return self._loop.put_coro(coro_worker, *args, **kwargs)
    
    def _tree_monitoring_state(self):
        return True, self._tree_monitoring_turned_on, None
    
    def _put_background_coro(self, coro_worker: AnyWorker, args, kwargs):
        try:
            coro: CoroWrapperBase = self._single_task_registration_or_immediate_processing_single_impl(coro_worker, *args, **kwargs)
            coro.is_background_coro = True
        except:
            return True, None, get_exception()

        return True, coro.coro_id, None
    
    def _task(self, coro_worker: AnyWorker, args: Tuple, kwargs: Dict, background: bool):
        try:
            coro: CoroWrapperBase = self._single_task_registration_or_immediate_processing_single_impl(coro_worker, *args, **kwargs)
            coro.is_background_coro = background
        except:
            return True, None, get_exception()

        return True, Task(coro), None

    def full_processing_iteration(self):
        if self._dead_coroutines:
            dead_coroutines_buff = self._dead_coroutines
            self._dead_coroutines = type(dead_coroutines_buff)()
            for dead_coro_id in dead_coroutines_buff:
                parent_coro_id = self._tree_parent_by_child.get(dead_coro_id, None)
                children = self._tree_children_by_parent.get(dead_coro_id, None)
                if parent_coro_id is None:
                    if children is not None:
                        for child_coro_id in children:
                            self._tree_parent_by_child.pop(child_coro_id, None)
                else:
                    sibblings = self._tree_children_by_parent[parent_coro_id]
                    sibblings.discard(dead_coro_id)
                    if children is not None:
                        sibblings.update(children)
                        for child_coro_id in children:
                            self._tree_parent_by_child[child_coro_id] = parent_coro_id

        if self.direct_requests:
            direct_requests_buff = self.direct_requests
            self.direct_requests = type(direct_requests_buff)()
            for coro_worker, args, kwargs in direct_requests_buff:
                try:
                    coro = self._loop.put_coro(coro_worker, *args, **kwargs)
                except:
                    ex_type, exception, tracback = get_exception_tripple()
                    if __debug__: dlog(ex_type, exception, tracback)
                    raise

        self.make_dead()
    
    def _add_direct_request(self, coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[None]:
        self.direct_requests.append((coro_worker, args, kwargs))
        self.make_live()
        return (False, None)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        self._dead_coroutines.add(coro.coro_id)
        self.make_live()
        return False
    
    def travers_through_all_children(self, coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None):
        t = KeyMultiValueTreeTraversal(self._tree_children_by_parent, None, handler, on_switched_to_stack_based_implementation)
        t(coro_id)
    
    def get_set_of_all_children(self, coro_id) -> Set[CoroID]:
        result = set()
        def handler(deep, parent, child: ValueExistence[CoroID], index):
            if child:
                result.add(child[1])
        
        self.travers_through_all_children(coro_id, handler)
        result.discard(coro_id)
        return result
    
    def travers_through_all_parents(self, coro_id, handler: Callable[[CoroID, CoroID], None], on_switched_to_stack_based_implementation: Optional[Callable]=None):
        t = KeyValueTreeTraversal(self._tree_parent_by_child, None, handler, on_switched_to_stack_based_implementation)
        t(coro_id)
    
    def get_set_of_all_parents(self, coro_id):
        result = set()
        def handler(deep, child, parent, index):
            if parent is not None:
                result.add(parent)
        
        self.travers_through_all_parents(coro_id, handler)
        result.discard(coro_id)
        return result
    
    def get_set_of_all_coros(self):
        result = set(self._tree_children_by_parent.keys())
        result.update(self._tree_parent_by_child.keys())
        return result

    def in_work(self) -> bool:
        result: bool = bool(self.direct_requests) or bool(self._dead_coroutines)
        return self.thrifty_in_work(result)


PutCoroRequest.default_service_type = PutCoro


def put_current_from_other_service(current_service: Service, coro_worker: AnyWorker, *args, **kwargs) -> CoroWrapperBase:
    put_coro: PutCoro = current_service._loop.get_service_instance(PutCoro)
    return put_coro.put_from_other_service(current_service.current_caller_coro_info.coro_id, coro_worker, *args, **kwargs)


def put_from_other_service(current_service: Service, coro_id: CoroID, coro_worker: AnyWorker, *args, **kwargs) -> CoroWrapperBase:
    put_coro: PutCoro = current_service._loop.get_service_instance(PutCoro)
    return put_coro.put_from_other_service(coro_id, coro_worker, *args, **kwargs)


def put_root_from_other_service(current_service: Service, coro_worker: AnyWorker, *args, **kwargs) -> CoroWrapperBase:
    put_coro: PutCoro = current_service._loop.get_service_instance(PutCoro)
    return put_coro.put_root_from_other_service(coro_worker, *args, **kwargs)


def put_coro_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    """_summary_
        context can be generated by one of the [interface_and_loop_with_backup_loop, get_interface_and_loop_with_backup_loop, interface_and_loop_with_explicit_loop, get_interface_and_loop_with_explicit_loop, interface_for_an_explicit_loop, get_interface_for_an_explicit_loop] functions from the cengal/parallel_execution/coroutines/coro_scheduler module
        
        An example:

        from cengal.parallel_execution.coroutines.coro_scheduler import get_interface_and_loop_with_explicit_loop, CoroSchedulerType, ExplicitWorker, Worker, CoroID
        from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import put_coro_to
        from typing import Optional, Union

        def my_func(loop: CoroSchedulerType, coro_worker: AnyWorker, a, b) -> Optional[CoroID]:
            resulting_coro_id: Optional[CoroID] = None
            try:
                result = put_coro_to(get_interface_and_loop_with_explicit_loop(loop), coro_worker, a, b)
                if result:
                    print(f'We are inside of the loop AND in the coroutine. Our requested coro already was created at this moment. Coro id: {result.value}')
                    resulting_coro_id = result.value
                else:
                    print('We are outside of the loop or not in the coroutine body. Our requested coroutine will be created in the near future.')
            except CoroSchedulerContextIsNotAvailable:
                print('We are outside of the loop AND no loop was selected as a Primary AND our given `loop` var is None)
            
            return resulting_coro_id
        
    Args:
        context (Tuple[Optional[CoroSchedulerType], Optional[Interface], bool]): _description_
        coro_worker (AnyWorker): _description_

    Returns:
        ValueExistence[Optional[CoroID]]: _description_
    """
    return make_request_to_service_with_context(context, PutCoro, coro_worker, *args, **kwargs)


def try_put_coro_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    """_summary_
        context can be generated by one of the [interface_and_loop_with_backup_loop, get_interface_and_loop_with_backup_loop, interface_and_loop_with_explicit_loop, get_interface_and_loop_with_explicit_loop, interface_for_an_explicit_loop, get_interface_for_an_explicit_loop] functions from the cengal/parallel_execution/coroutines/coro_scheduler module
        
        An example:

        from cengal.parallel_execution.coroutines.coro_scheduler import get_interface_and_loop_with_explicit_loop, CoroSchedulerType, ExplicitWorker, Worker, CoroID
        from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import try_put_coro_to
        from typing import Optional, Union

        def my_func(loop: CoroSchedulerType, coro_worker: AnyWorker, a, b) -> Optional[CoroID]:
            resulting_coro_id: Optional[CoroID] = None
            result = put_coro_to(get_interface_and_loop_with_explicit_loop(loop), coro_worker, a, b)
            if result:
                print(f'We are inside of the loop AND in the coroutine. Our requested coro already was created at this moment. Coro id: {result.value}')
                resulting_coro_id = result.value
            else:
                print('There are two possibilities:)
                print('1) We are outside of the loop or not in the coroutine body. Our requested coroutine will be created in the near future.')
                print('2) We are outside of the loop AND no loop was selected as a Primary AND our given `loop` var is None. Our requested coroutine will NOT be created at all.)
            
            return resulting_coro_id
        
    Args:
        context (Tuple[Optional[CoroSchedulerType], Optional[Interface], bool]): _description_
        coro_worker (AnyWorker): _description_

    Returns:
        ValueExistence[Optional[CoroID]]: _description_
    """
    return try_make_request_to_service_with_context(context, PutCoro, coro_worker, *args, **kwargs)


async def aput_coro_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return await amake_request_to_service_with_context(context, PutCoro, coro_worker, *args, **kwargs)


async def atry_put_coro_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service_with_context(context, PutCoro, coro_worker, *args, **kwargs)


def put_coro(coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return make_request_to_service(PutCoro, coro_worker, *args, **kwargs)


def try_put_coro(coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return try_make_request_to_service(PutCoro, coro_worker, *args, **kwargs)


async def aput_coro(coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[CoroID]:
    return await amake_request_to_service(PutCoro, coro_worker, *args, **kwargs)


async def atry_put_coro(coro_worker: AnyWorker, *args, **kwargs) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service(PutCoro, coro_worker, *args, **kwargs)


# ==================================================


def travers_through_all_coro_children_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[CoroID]:
    put_coro: PutCoro = service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    return put_coro.travers_through_all_children(coro_id, handler, on_switched_to_stack_based_implementation)


def try_travers_through_all_coro_children_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[Optional[CoroID]]:
    put_coro: PutCoro = get_service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    if put_coro is not None:
        return put_coro.travers_through_all_children(coro_id, handler, on_switched_to_stack_based_implementation)


def travers_through_all_coro_children(coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[CoroID]:
    return travers_through_all_coro_children_on(get_available_coro_scheduler(), coro_id, handler, on_switched_to_stack_based_implementation)


def try_travers_through_all_coro_children(coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[Optional[CoroID]]:
    return try_travers_through_all_coro_children_on(get_available_coro_scheduler(), coro_id, handler, on_switched_to_stack_based_implementation)


# ==================================================


def get_set_of_all_coro_children_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id) -> Set[CoroID]:
    put_coro: PutCoro = service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    return put_coro.get_set_of_all_children(coro_id)


def try_get_set_of_all_coro_children_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id) -> Optional[Set[CoroID]]:
    put_coro: PutCoro = get_service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    if put_coro is not None:
        return put_coro.get_set_of_all_children(coro_id)


def get_set_of_all_coro_children(coro_id) -> Set[CoroID]:
    return get_set_of_all_coro_children_on(get_available_coro_scheduler(), coro_id)


def try_get_set_of_all_coro_children(coro_id) -> Optional[Set[CoroID]]:
    return try_get_set_of_all_coro_children_on(get_available_coro_scheduler(), coro_id)


# ==================================================


def travers_through_all_coro_parents_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[CoroID]:
    put_coro: PutCoro = service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    return put_coro.travers_through_all_parents(coro_id, handler, on_switched_to_stack_based_implementation)


def try_travers_through_all_coro_parents_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[Optional[CoroID]]:
    put_coro: PutCoro = get_service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    if put_coro is not None:
        return put_coro.travers_through_all_parents(coro_id, handler, on_switched_to_stack_based_implementation)


def travers_through_all_coro_parents(coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[CoroID]:
    return travers_through_all_coro_parents_on(get_available_coro_scheduler(), coro_id, handler, on_switched_to_stack_based_implementation)


def try_travers_through_all_coro_parents(coro_id, handler: Callable[[int, CoroID, CoroID, int], None], on_switched_to_stack_based_implementation: Optional[Callable]=None) -> ValueExistence[Optional[CoroID]]:
    return try_travers_through_all_coro_parents_on(get_available_coro_scheduler(), coro_id, handler, on_switched_to_stack_based_implementation)


# ==================================================


def get_set_of_all_coro_parents_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id) -> ValueExistence[CoroID]:
    put_coro: PutCoro = service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    return put_coro.get_set_of_all_parents(coro_id)


def try_get_set_of_all_coro_parents_on(prioritized_coro_scheduler: Optional[CoroSchedulerType], coro_id) -> ValueExistence[Optional[CoroID]]:
    put_coro: PutCoro = get_service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    if put_coro is not None:
        return put_coro.get_set_of_all_parents(coro_id)


def get_set_of_all_coro_parents(coro_id) -> ValueExistence[CoroID]:
    return get_set_of_all_coro_parents_on(get_available_coro_scheduler(), coro_id)


def try_get_set_of_all_coro_parents(coro_id) -> ValueExistence[Optional[CoroID]]:
    return try_get_set_of_all_coro_parents_on(get_available_coro_scheduler(), coro_id)


# ==================================================


def get_set_of_all_coros_on(prioritized_coro_scheduler: Optional[CoroSchedulerType]) -> ValueExistence[CoroID]:
    put_coro: PutCoro = service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    return put_coro.get_set_of_all_coros()


def try_get_set_of_all_coros_on(prioritized_coro_scheduler: Optional[CoroSchedulerType]) -> ValueExistence[Optional[CoroID]]:
    put_coro: PutCoro = get_service_with_explicit_loop(PutCoro, prioritized_coro_scheduler)
    if put_coro is not None:
        return put_coro.get_set_of_all_coros()


def get_set_of_all_coros() -> ValueExistence[CoroID]:
    return get_set_of_all_coros_on(get_available_coro_scheduler())


def try_get_set_of_all_coros() -> ValueExistence[Optional[CoroID]]:
    return try_get_set_of_all_coros_on(get_available_coro_scheduler())



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__author__ = 'Mikhail Butenko <gtalk@mikhail-butenko.in.ua>'


# import TagDB
# from TagDB import USUAL_TREE_TYPE as USUAL_TREE_TYPE
# from TagDB import SMART_TREE_TYPE as SMART_TREE_TYPE
# from TagDB import FULL_TREE_TYPE as FULL_TREE_TYPE
# from TagDB import PLAIN_PSEUDO_TREE_TYPE as PLAIN_PSEUDO_TREE_TYPE
from .TagDB import *
import copy
from cengal.web_tools.request_cache import RequestCache
import time


class tag_db_interface:

    def __init__(self):
        super().__init__()
        self.__db = TagDB()
        self.cache = RequestCache(300000)

    def get_db(self):
        return self.__db

    def add_item(self, binItem, binTags):
        binItem = copy.deepcopy(binItem)
        binTags = copy.deepcopy(binTags)

        result = self.__db.add_item(binItem, binTags)

        result = copy.deepcopy(result)
        self.cache.clear()
        return result

    def remove_item(self, binTags, binItem):
        binTags = copy.deepcopy(binTags)
        binItem = copy.deepcopy(binItem)

        result = self.__db.remove_item(binItem, binTags)

        result = copy.deepcopy(result)
        self.cache.clear()
        return result

    def get_items(self, binTags, treeType=USUAL_TREE_TYPE):
        binTags = copy.deepcopy(binTags)
        treeType = copy.deepcopy(treeType)

        # request = ('get_items', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
        request = ('get_items', frozenset(binTags), treeType)
        result = self.cache.try_to_get_data_for_request(request)
        if result is None:
            result = self.__db.get_items_from_tags(binTags, treeType=treeType)
            self.cache.put_new_request(request, result)

        result = copy.deepcopy(result)
        return result

    def get_tags(self, binTags, treeType=USUAL_TREE_TYPE):
        binTags = copy.deepcopy(binTags)
        treeType = copy.deepcopy(treeType)

        # request = ('get_tags', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
        request = ('get_tags', frozenset(binTags), treeType)
        result = self.cache.try_to_get_data_for_request(request)
        if result is None:
            result = self.__db.get_tags_from_tags(binTags, treeType=treeType)
            self.cache.put_new_request(request, result)

        result = copy.deepcopy(result)
        return result

    def get_all(self, binTags, treeType=USUAL_TREE_TYPE):
        binTags = copy.deepcopy(binTags)
        treeType = copy.deepcopy(treeType)

        # request = ('get_all', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
        request = ('get_all', frozenset(binTags), treeType)
        result = self.cache.try_to_get_data_for_request(request)
        if result is None:
            result = self.__db.get_all_from_tags(binTags, treeType=treeType)
            self.cache.put_new_request(request, result)

        result = copy.deepcopy(result)
        return result

    def get_tags_for_a_smart_redirection(self, binTags):
        binTags = copy.deepcopy(binTags)

        # request = ('is_smart_redirection_for_a_tag_path_reduction_needed'
        #            , tuple(self.__db.sort_tag_list_by_hash(binTags)))
        request = ('is_smart_redirection_for_a_tag_path_reduction_needed', frozenset(binTags))
        result = self.cache.try_to_get_data_for_request(request)
        if result is None:
            result = self.__db.get_tags_for_a_smart_redirection(binTags)
            self.cache.put_new_request(request, result)

        result = copy.deepcopy(result)
        return result

    def fill_cache(self, max_answer_time_in_miliseconds=500, path=list()):
        current_path = list(path)
        startTime = time.time()

        tags_and_items_inside_curent_path = self.get_all(current_path, SMART_TREE_TYPE)
        internal_tags = tags_and_items_inside_curent_path[0]
        self.get_items(current_path, PLAIN_PSEUDO_TREE_TYPE)

        endTime = time.time()
        resultTime = endTime - startTime
        if (resultTime*1000) >= max_answer_time_in_miliseconds:
            for additional_tag in internal_tags:
                new_path = current_path + [additional_tag]
                self.fill_cache(max_answer_time_in_miliseconds, new_path)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['BindableToType', 'AlreadyBoundError', 'NotBoundYetError']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from typing import Tuple, Any, Optional
from inspect import isclass


class AlreadyBoundError(Exception):
    pass


class NotBoundYetError(Exception):
    pass


class BindableToType:
    """ Bindable to type or to type of given object

    Example:

        class MyGlobalData(BindableToType):
            def __init__(self) -> None:
                super().__init__()
                self.data = 2


        class MyStructure:
            global_data = MyGlobalData()
            def __init__(self):
                self.index = 3


        print(f'{MyStructure.global_data.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data')

        ms = MyStructure()

        print(f'{ms.global_data.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data')

        MyGlobalData().bind(MyStructure, 'global_data_2')

        print(f'{ms.global_data_2.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data')
        print(f'{MyStructure.global_data_2.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data')

        MyGlobalData().bind(ms, 'global_data_3')

        print(f'{ms.global_data_3.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data_2')
        print(f'{MyStructure.global_data_3.bound_to() = }')  # (<class '__main__.MyStructure'>, 'global_data_2')
    """    
    def __init__(self) -> None:
        self._bound_to_owner_type = None
        self._bound_to_name = None
    
    def on_bind(self, owner_type, name):
        pass
    
    def on_unbind(self, owner_type, name):
        pass

    def __set_name__(self, owner_type, name):
        self._bound_to_owner_type = owner_type
        self._bound_to_name = name
        self.on_bind(owner_type, name)
    
    def bind(self, owner_type_or_obj, name):
        if self._bound_to_owner_type is not None:
            raise AlreadyBoundError
        
        if isclass(owner_type_or_obj):
            owner_class = owner_type_or_obj
        else:
            owner_class = type(owner_type_or_obj)
        
        setattr(owner_class, name, self)
        self.__set_name__(owner_class, name)
    
    def unbind(self) -> Tuple[Any, str]:
        if self._bound_to_owner_type is None:
            raise NotBoundYetError
        
        delattr(self._bound_to_owner_type, self._bound_to_name)
        result = (self._bound_to_owner_type, self._bound_to_name)
        self._bound_to_owner_type = None
        self._bound_to_name = None
        self.on_unbind(self._bound_to_owner_type, self._bound_to_name)
        return result
    
    def bound(self) -> bool:
        return self._bound_to_owner_type is not None
    
    def bound_to(self) -> Optional[Tuple[Any, str]]:
        result = None
        if self.bound():
            result = (self._bound_to_owner_type, self._bound_to_name)
        
        return result
    
    def force_bind(self, owner_type_or_obj, name) -> Optional[Tuple[Any, str]]:
        result = None
        if self.bound():
            result = self.unbind()
        
        self.bind(owner_type_or_obj, name)
        return result
